import argparse
from email import parser
import importlib.util
import itertools
import json
import os
import random
import re
import shutil
import socket
import subprocess  # nosec B404
import sys
from abc import ABC, abstractmethod
from contextlib import closing
from enum import Enum
from pathlib import Path
from typing import Any, TextIO, cast

import numpy as np
import psutil
import pynvml
import requests
import torch
import torch.distributed as dist
import torch.nn.functional as F  # noqa: N812
from huggingface_hub import HfApi, hf_hub_download
from huggingface_hub.hf_api import RepoFolder
from packaging.requirements import Requirement
from ruamel.yaml import YAML
from torch.utils.data import DataLoader
from torch.utils.tensorboard.writer import SummaryWriter
from tqdm import tqdm
from functools import partial
import tempfile

from konfai import (
    config_file,
    cuda_visible_devices,
    evaluations_directory,
    konfai_state,
    predictions_directory,
    statistics_directory,
)


def description(model, model_ema=None, show_memory: bool = True, train: bool = True) -> str:
    def loss_desc(model):
        return (
            "("
            + " ".join(
                [
                    f"{name}({(network.optimizer.param_groups[0]['lr'] if network.optimizer else 0):.6f}) : "
                    + " ".join(
                        f"{k.split(':')[-1]}({w:.2f}) : {v:.6f}"
                        for (k, v), w in zip(
                            network.measure.get_last_values().items(), network.measure.get_last_weights().values()
                        )
                    )
                    for name, network in model.module.get_networks().items()
                    if network.measure is not None
                ]
            )
            + ")"
        )

    model_loss_desc = loss_desc(model)
    result = ""
    if len(model_loss_desc) > 2:
        result += f"Loss {model_loss_desc} "
    if model_ema is not None:
        model_ema_loss_desc = loss_desc(model_ema)
        if len(model_ema_loss_desc) > 2:
            result += f"Loss EMA {model_ema_loss_desc} "
    result += gpu_info()
    if show_memory:
        result += f" | {get_memory_info()}"
    return result


def get_module(classpath: str, default_classpath: str) -> tuple[str, str]:
    if len(classpath.split(":")) > 1:
        module = ".".join(classpath.split(":")[:-1])
        name = classpath.split(":")[-1]
    else:
        module = (
            default_classpath + ("." if len(classpath.split(".")) > 2 else "") + ".".join(classpath.split(".")[:-1])
        )
        name = classpath.split(".")[-1]
    return module, name.split("/")[0]


def get_cpu_info() -> str:
    return f"CPU ({psutil.cpu_percent(interval=0.5):.2f} %)"


def get_memory_info() -> str:
    return f"Memory ({psutil.virtual_memory().used / 2**30:.2f}G ({psutil.virtual_memory().percent:.2f} %))"


def get_memory() -> float:
    return psutil.virtual_memory().used / 2**30


def memory_forecast(memory_init: float, size: float) -> str:
    current_memory = get_memory()
    forecast = memory_init + ((current_memory - memory_init) * size)
    return f"Memory forecast ({forecast:.2f}G ({forecast / (psutil.virtual_memory().total / 2**30) * 100:.2f} %))"


def gpu_info() -> str:
    if cuda_visible_devices() == "":
        return ""

    devices = [int(i) for i in cuda_visible_devices().split(",")]
    device = devices[0]

    if device < pynvml.nvmlDeviceGetCount():
        handle = pynvml.nvmlDeviceGetHandleByIndex(device)
        memory = pynvml.nvmlDeviceGetMemoryInfo(handle)
    else:
        return ""
    node_name = "Node: {} " + os.environ["SLURMD_NODENAME"] if "SLURMD_NODENAME" in os.environ else ""
    return f"{node_name}GPU({devices}) Memory GPU ({memory.used / 1e9:.2f}G ({memory.used / memory.total * 100:.2f} %))"


def get_max_gpu_memory(device: int | torch.device) -> float:
    if isinstance(device, torch.device):
        if str(device).startswith("cuda:"):
            device = int(str(device).replace("cuda:", ""))
        else:
            return 0
    device = [int(i) for i in cuda_visible_devices().split(",")][device]
    if device < pynvml.nvmlDeviceGetCount():
        handle = pynvml.nvmlDeviceGetHandleByIndex(device)
        memory = pynvml.nvmlDeviceGetMemoryInfo(handle)
    else:
        return 0
    return float(memory.total) / (10**9)


def get_gpu_memory(device: int | torch.device) -> float:
    if isinstance(device, torch.device):
        if str(device).startswith("cuda:"):
            device = int(str(device).replace("cuda:", ""))
        else:
            return 0
    device = [int(i) for i in cuda_visible_devices().split(",")][device]
    if device < pynvml.nvmlDeviceGetCount():
        handle = pynvml.nvmlDeviceGetHandleByIndex(device)
        memory = pynvml.nvmlDeviceGetMemoryInfo(handle)
    else:
        return 0
    return float(memory.used) / (10**9)


class NeedDevice:

    def __init__(self) -> None:
        super().__init__()
        self.device: torch.device

    def to(self, device: int):
        self.device = get_device(device)


def get_device(device: int):
    return device if torch.cuda.is_available() and device >= 0 else torch.device("cpu")


class State(Enum):
    TRAIN = "TRAIN"
    RESUME = "RESUME"
    TRANSFER_LEARNING = "TRANSFER_LEARNING"
    FINE_TUNING = "FINE_TUNING"
    PREDICTION = "PREDICTION"
    EVALUATION = "EVALUATION"

    def __str__(self) -> str:
        return self.value


def get_patch_slices_from_nb_patch_per_dim(
    patch_size_tmp: list[int],
    nb_patch_per_dim: list[tuple[int, bool]],
    overlap: int | None,
) -> list[tuple[slice, ...]]:
    patch_slices = []
    slices: list[list[slice]] = []
    if overlap is None:
        overlap = 0
    patch_size = []
    i = 0
    for nb in nb_patch_per_dim:
        if nb[1]:
            patch_size.append(1)
        else:
            patch_size.append(patch_size_tmp[i])
            i += 1

    for dim, nb in enumerate(nb_patch_per_dim):
        slices.append([])
        for index in range(nb[0]):
            start = (patch_size[dim] - overlap) * index
            end = start + patch_size[dim]
            slices[dim].append(slice(start, end))
    for chunk in itertools.product(*slices):
        patch_slices.append(tuple(chunk))
    return patch_slices


def get_patch_slices_from_shape(
    patch_size: list[int], shape: list[int], overlap_tmp: int | None
) -> tuple[list[tuple[slice, ...]], list[tuple[int, bool]]]:
    if len(shape) != len(patch_size):
        raise DatasetManagerError(
            f"Dimension mismatch: 'patch_size' has {len(patch_size)} dimensions, but 'shape' has {len(shape)}.",
            f"patch_size: {patch_size}",
            f"shape: {shape}",
            "Both must have the same number of dimensions (e.g., 3D patch for 3D volume).",
        )
    patch_slices = []
    nb_patch_per_dim = []
    slices: list[list[slice]] = []
    if overlap_tmp is None:
        size = [np.ceil(a / b) for a, b in zip(shape, patch_size)]
        tmp = np.zeros(len(size), dtype=np.int_)
        for i, s in enumerate(size):
            if s > 1:
                tmp[i] = np.mod(patch_size[i] - np.mod(shape[i], patch_size[i]), patch_size[i]) // (size[i] - 1)
        overlap = tmp
    else:
        overlap = [overlap_tmp if size > 1 else 0 for size in patch_size]

    for dim in range(len(shape)):
        if overlap[dim] >= patch_size[dim]:
            raise ValueError(
                f"Overlap must be less than patch size, got overlap={overlap[dim]}",
                f" ≥ patch_size={patch_size[dim]} at dim={dim}",
            )

    for dim in range(len(shape)):
        slices.append([])
        index = 0
        while True:
            start = (patch_size[dim] - overlap[dim]) * index

            end = start + patch_size[dim]
            if end >= shape[dim]:
                end = shape[dim]
                slices[dim].append(slice(start, end))
                break
            slices[dim].append(slice(start, end))
            index += 1
        nb_patch_per_dim.append((index + 1, patch_size[dim] == 1))

    for chunk in itertools.product(*slices):
        patch_slices.append(tuple(chunk))

    return patch_slices, nb_patch_per_dim


def _log_signal_format(array: np.ndarray) -> dict[str, np.ndarray]:
    return {str(i): channel for i, channel in enumerate(array)}


def _log_image_format(array: np.ndarray) -> np.ndarray:
    if len(array.shape) == 2:
        array = np.expand_dims(array, axis=0)

    if len(array.shape) == 3 and array.shape[0] != 1:
        array = np.expand_dims(array, axis=0)
    if len(array.shape) == 4:
        array = array[:, array.shape[1] // 2]

    array = array.astype(float)
    b = -np.min(array)
    if (np.max(array) + b) > 0:
        return (array + b) / (np.max(array) + b)
    else:
        return 0 * array


def _log_images_format(array: np.ndarray) -> np.ndarray:
    result = []
    for n in range(array.shape[0]):
        result.append(_log_image_format(array[n]))
    result = np.stack(result, axis=0)
    return result


def _log_video_format(array: np.ndarray) -> np.ndarray:
    result_list = []
    for t in range(array.shape[1]):
        result_list.append(_log_images_format(array[:, t, ...]))
    result = np.stack(result_list, axis=1)

    nb_channel = result.shape[2]
    if nb_channel < 3:
        channel_split = [result[:, :, 0, ...] for i in range(3)]
    else:
        channel_split = np.split(result, 3, axis=0)
    array = np.zeros((result.shape[0], result.shape[1], 3, *list(result.shape[3:])))
    for i, channels in enumerate(channel_split):
        array[:, :, i] = np.mean(channels, axis=0)
    return array


class DataLog(Enum):
    SIGNAL = "SIGNAL"
    IMAGE = "IMAGE"
    IMAGES = "IMAGES"
    VIDEO = "VIDEO"
    AUDIO = "AUDIO"

    def __call__(self, tb: SummaryWriter, name: str, layer: torch.Tensor, it: int):
        if self == DataLog.SIGNAL:
            return [
                tb.add_scalars(name, _log_signal_format(layer[b, :, 0]), layer.shape[0] * it + b)
                for b in range(layer.shape[0])
            ]
        elif self == DataLog.IMAGE:
            return tb.add_image(name, _log_image_format(layer[0]), it)
        elif self == DataLog.IMAGES:
            return tb.add_images(name, _log_images_format(layer), it)
        elif self == DataLog.VIDEO:
            return tb.add_video(name, _log_video_format(layer), it)
        elif self == DataLog.AUDIO:
            return tb.add_audio(name, _log_image_format(layer), it)
        else:
            raise ValueError(f"Unsupported DataLog type: {self}")


class Log:
    def __init__(self, name: str, rank: int) -> None:
        if konfai_state() == "PREDICTION":
            path = predictions_directory()
        elif konfai_state() == "EVALUATION":
            path = evaluations_directory()
        else:
            path = statistics_directory()

        self.verbose = os.environ.get("KONFAI_VERBOSE", "True") == "True"
        self.log_path = os.path.join(path, name)
        os.makedirs(self.log_path, exist_ok=True)
        self.rank = rank
        self.file = open(os.path.join(self.log_path, f"log_{rank}.txt"), "w", buffering=1)
        self.stdout_bak = sys.stdout
        self.stderr_bak = sys.stderr
        self._buffered_line = ""

    def __enter__(self):
        self.file.__enter__()
        sys.stdout = cast(TextIO, self)
        sys.stderr = cast(TextIO, self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.__exit__(exc_type, exc_val, exc_tb)
        sys.stdout = self.stdout_bak
        sys.stderr = self.stderr_bak

    def write(self, msg: str):
        if not msg:
            return

        ansi_escape = re.compile(r"\x1b\[[0-9;]*[a-zA-Z]")
        msg_clean = ansi_escape.sub("", msg)
        if "\r" in msg_clean or "[A" in msg:
            msg_clean = msg_clean.split("\r")[-1].strip()
            self._buffered_line = msg_clean
        else:
            self._buffered_line = msg_clean.strip()

        if self._buffered_line:
            self.file.write(self._buffered_line + "\n")
            self.file.flush()
        if self.verbose and (self.rank == 0 or "KONFAI_CLUSTER" in os.environ):
            if sys.__stdout__ is not None:
                sys.__stdout__.write(msg)
                sys.__stdout__.flush()

    def flush(self):
        self.file.flush()

    def isatty(self):
        return False

    def fileno(self):
        if sys.__stdout__ is None:
            raise RuntimeError("sys.__stdout__ is None, cannot get fileno")
        return sys.__stdout__.fileno()


class TensorBoard:

    def __init__(self, name: str) -> None:
        self.process: subprocess.Popen | None = None
        self.name = name

    def __enter__(self):
        if "KONFAI_TENSORBOARD_PORT" in os.environ:
            tensorboard_exe = shutil.which("tensorboard")
            if tensorboard_exe is None:
                raise RuntimeError("TensorBoard executable not found in PATH.")

            logdir = (
                predictions_directory() if konfai_state() == "PREDICTION" else statistics_directory() + self.name + "/"
            )

            port = os.environ.get("KONFAI_TENSORBOARD_PORT")
            if not port or not port.isdigit():
                raise ValueError("Invalid or missing KONFAI_TENSORBOARD_PORT.")

            command = [
                tensorboard_exe,
                "--logdir",
                logdir,
                "--port",
                port,
                "--bind_all",
            ]
            self.process = subprocess.Popen(command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # nosec B603
            try:
                s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                s.connect(("10.255.255.255", 1))
                ip = s.getsockname()[0]
            except Exception:
                ip = "127.0.0.1"
            finally:
                s.close()
            print(f"Tensorboard : http://{ip}:{os.environ['KONFAI_TENSORBOARD_PORT']}/")
        return self

    def __exit__(self, exc_type, value, traceback):
        if self.process is not None:
            self.process.terminate()
            self.process.wait()


class DistributedObject(ABC):

    def __init__(self, name: str) -> None:
        self.port = find_free_port()
        self.dataloader: list[list[DataLoader]]
        self.manual_seed: int | None = None
        self.name = name
        self.size = 1

    @abstractmethod
    def setup(self, world_size: int):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, value, traceback):
        cleanup()

    @abstractmethod
    def run_process(
        self,
        world_size: int,
        global_rank: int,
        local_rank: int,
        dataloaders: list[DataLoader],
    ):
        pass

    @staticmethod
    def get_measure(
        world_size: int,
        global_rank: int,
        gpu: int,
        models: dict[str, torch.nn.Module],
        n: int,
    ) -> dict[str, tuple[dict[str, tuple[float, float]], dict[str, tuple[float, float]]]]:
        data = {}
        for label, model in models.items():
            for name, network in model.get_networks().items():
                if network.measure is not None:
                    data[f"{name}{label}"] = (
                        network.measure.format_loss(True, n),
                        network.measure.format_loss(False, n),
                    )
        outputs = synchronize_data(world_size, gpu, data)
        result: dict[str, tuple[dict[str, tuple[float, float]], dict[str, tuple[float, float]]]] = {}
        if global_rank == 0:
            for output in outputs:
                for k, v in output.items():
                    for t in range(len(v)):
                        for u, n in v[t].items():
                            if k not in result:
                                result[k] = ({}, {})
                            if u not in result[k][t]:
                                result[k][t][u] = (n[0], 0)  # type: ignore[index]
                            result[k][t][u] = (
                                result[k][t][u][0],
                                result[k][t][u][1] + n[1] / world_size,  # type: ignore[index]
                            )
        return result

    def __call__(self, rank: int | None = None) -> None:
        world_size = len(self.dataloader)
        global_rank, local_rank = setup_gpu(world_size, self.port, rank)
        if global_rank is None or local_rank is None:
            return
        with Log(self.name, global_rank):
            if torch.cuda.is_available():
                pynvml.nvmlInit()
            if self.manual_seed is not None:
                np.random.seed(self.manual_seed * world_size + global_rank)
                random.seed(self.manual_seed * world_size + global_rank)
                torch.manual_seed(self.manual_seed * world_size + global_rank)
            torch.backends.cudnn.benchmark = self.manual_seed is None
            torch.backends.cudnn.deterministic = self.manual_seed is not None
            # torch.backends.cuda.matmul.fp32_precision = "tf32"
            # torch.backends.cudnn.conv.fp32_precision = "tf32"
            dataloaders = self.dataloader[global_rank]
            if torch.cuda.is_available():
                torch.cuda.set_device(local_rank)
            try:
                self.run_process(world_size, global_rank, local_rank, dataloaders)
            finally:
                cleanup()
                if torch.cuda.is_available():
                    pynvml.nvmlShutdown()


def match_supported(file: Path) -> bool:
    lower = file.name.lower()
    return any(lower.endswith("." + ext) for ext in SUPPORTED_EXTENSIONS)


def list_supported_files(path: str):
    path = Path(path)
    if path.is_file():
        return [path] if match_supported(path) else []

    files = []
    for f in path.rglob("*"):
        if f.is_file() and match_supported(f):
            files.append(f)
    return files

def setup_apps(parser: argparse.ArgumentParser, user_dir: Path, tmp_dir_default: Path) -> tuple[DistributedObject, str]:
    def get_path(path: Path) -> Path:
        if path.is_absolute():
            result_path = path
        else:
            result_path = user_dir / path
        return result_path
    parser = argparse.ArgumentParser(prog="konfai-apps", description="KonfAI Apps – Apps for Medical AI Models")

    subparsers = parser.add_subparsers(dest="command", required=True)

    # -----------------
    # 1) INFERENCE
    # -----------------
    infer_p = subparsers.add_parser("infer", help="Run inference using a KonfAI App.")

    infer_p.add_argument("app", type=str, help="KonfAI App name")
    infer_p.add_argument(
        "-i", "--input", type=lambda p: get_path(Path(p)), required=True, help="Input path: either a single volume file OR a dataset directory"
    )
    infer_p.add_argument("-o", "--output", type=lambda p: get_path(Path(p)), default="./Predictions", help="Optional output volume path")
    infer_p.add_argument("--ensemble", type=int, default=0, help="Size of model ensemble")
    infer_p.add_argument("--tta", type=int, default=0, help="Number of Test-Time Augmentations")
    infer_p.add_argument("--mc", type=int, default=0, help="Monte Carlo dropout samples")
    infer_p.add_argument(
        "-g",
        "--gpu",
        type=str,
        default=(os.environ["CUDA_VISIBLE_DEVICES"] if "CUDA_VISIBLE_DEVICES" in os.environ else ""),
        help="GPU list (e.g. '0' or '0,1'). Leave empty for CPU.",
    )
    infer_p.add_argument("--cpu", type=int, default=1, help="Number of CPU cores to use when --gpu is empty.")
    infer_p.add_argument("--tmp_dir", type=lambda p: Path(p).absolute(), default=tmp_dir_default, help="Use a temporary directory.")
    infer_p.add_argument("--quiet", action="store_false", help="Suppress console output.")

    # -----------------
    # 2) EVALUATION
    # -----------------
    eval_p = subparsers.add_parser("eval", help="Evaluate a KonfAI App using ground-truth labels.")
    eval_p.add_argument("app", type=str, help="KonfAI App name")

    eval_p.add_argument(
        "-i", "--input", type=lambda p: get_path(Path(p)), required=True, help="Input path: either a single volume file OR a dataset directory"
    )

    eval_p.add_argument("--gt", type=lambda p: get_path(Path(p)), required=True, help="Ground-truth path")
    eval_p.add_argument("--mask", type=lambda p: get_path(Path(p)), help="Optional evaluation mask path")
    eval_p.add_argument("-o", "--output", type=lambda p: get_path(Path(p)), default="./Evaluations", help="Optional output volume path")
    eval_p.add_argument("--tmp_dir", type=lambda p: Path(p).absolute(), default=tmp_dir_default, help="Use a temporary directory.")


    eval_p.add_argument(
        "-g",
        "--gpu",
        type=str,
        default=(os.environ["CUDA_VISIBLE_DEVICES"] if "CUDA_VISIBLE_DEVICES" in os.environ else ""),
        help="GPU list (e.g. '0' or '0,1'). Leave empty for CPU.",
    )
    eval_p.add_argument("--cpu", type=int, default=1, help="Number of CPU cores to use when --gpu is empty.")
    eval_p.add_argument("--quiet", action="store_false", help="Suppress console output.")

    # -----------------
    # 3) UNCERTAINTY
    # -----------------
    unc_p = subparsers.add_parser("uncertainty", help="Compute model uncertainty for a KonfAI App.")
    unc_p.add_argument("app", type=str, help="KonfAI App name")

    unc_p.add_argument(
        "-i",
        "--input",
        type=lambda p: get_path(Path(p)),
        required=True,
        help=("Prediction stack: either a single multi-sample prediction file "),
    )
    
    unc_p.add_argument("-o", "--output", type=lambda p: get_path(Path(p)), default="./Evaluations", help="Optional output volume path")
    unc_p.add_argument("--tmp_dir", type=lambda p: Path(p).absolute(), default=tmp_dir_default, help="Use a temporary directory.")


    unc_p.add_argument(
        "-g",
        "--gpu",
        type=str,
        default=(os.environ["CUDA_VISIBLE_DEVICES"] if "CUDA_VISIBLE_DEVICES" in os.environ else ""),
        help="GPU list (e.g. '0' or '0,1'). Leave empty for CPU.",
    )
    unc_p.add_argument("--cpu", type=int, default=1, help="Number of CPU cores to use when --gpu is empty.")
    unc_p.add_argument("--quiet", action="store_false", help="Suppress console output.")

    # -----------------
    # 4) Pipeline
    # -----------------
    pipe_p = subparsers.add_parser(
        "pipeline", help="Run inference and optionally evaluation and uncertainty in a single command."
    )

    pipe_p.add_argument("app", type=str, help="KonfAI App name")

    pipe_p.add_argument("-i", "--input", type=str, required=True, help="Input path: volume file or directory.")

    pipe_p.add_argument("--mc", type=int, default=0, help="Number of Monte Carlo dropout samples.")

    pipe_p.add_argument("--tta", type=int, default=0, help="Number of Test-Time Augmentations.")

    pipe_p.add_argument("--ensemble", type=int, default=0, help="Number of models in ensemble.")

    # optional eval
    pipe_p.add_argument("--with-eval", action="store_true", help="Also run evaluation (requires --gt).")

    pipe_p.add_argument("-g", "--gt", type=str, help="Ground-truth path (required when --with-eval is set).")

    # optional uncertainty
    pipe_p.add_argument("--with-uncertainty", action="store_true", help="Also estimate uncertainty using sampling.")

    pipe_p.add_argument("-o", "--output", type=str, help="Output prediction volume (mean prediction).")

    # -----------------
    # 5) FINE-TUNE
    # -----------------
    ft_p = subparsers.add_parser("fine-tune", help="Fine-tune a KonfAI App on a dataset.")
    ft_p.add_argument("app", type=str, help="KonfAI App name")

    ft_p.add_argument("-d", "--dataset", type=str, required=True, help="Path to training dataset")
    ft_p.add_argument("--epochs", type=int, default=10, help="Number of fine-tuning epochs")
    
    parser.add_argument("--version", action="version", version=importlib.metadata.version("IMPACT-Synth-KonfAI"))

    args = parser.parse_args()

    os.environ["CUDA_VISIBLE_DEVICES"] = str(args.gpu)
    os.environ["KONFAI_NB_CORES"] = str(args.cpu)
    os.environ["KONFAI_WORKERS"] = str(0)

    os.environ["KONFAI_MODELS_DIRECTORY"] = "./Models/"
    os.environ["KONFAI_CHECKPOINTS_DIRECTORY"] = "./Checkpoints/"
    os.environ["KONFAI_PREDICTIONS_DIRECTORY"] = str(args.output)+"/"
    os.environ["KONFAI_EVALUATIONS_DIRECTORY"] = str(args.output)+"/"
    os.environ["KONFAI_STATISTICS_DIRECTORY"] = "./Statistics/"
    os.environ["KONFAI_SETUPS_DIRECTORY"] = "./Setups/"
    os.environ["KONFAI_OVERWRITE"] = str(True)
    os.environ["KONFAI_CONFIG_MODE"] = "Done"

    os.environ["KONFAI_VERBOSE"] = str(args.quiet)

    model_hf = ModelHF(args.app)

    os.makedirs(args.tmp_dir, exist_ok=True)
    os.chdir(str(args.tmp_dir))
    if str(args.tmp_dir) not in sys.path:
        sys.path.insert(0, str(args.tmp_dir))

    if args.command != "fine-tune":
        from konfai.utils.dataset import Dataset
        import SimpleITK as sitk
        dataset = Dataset("Dataset", "mha")
        for idx, file in enumerate(list_supported_files(args.input)):
            print("jdckez")
            dataset.write("Volume", "P{:03d}".format(idx), sitk.ReadImage(str(file)))

    if args.command == "infer":
        models_path = model_hf.install_inference(args.tta, args.ensemble, args.mc)

        os.environ["KONFAI_ROOT"] = "Predictor"
        os.environ["KONFAI_config_file"] = "Prediction.yml"
        os.environ["KONFAI_MODEL"] = ":".join(models_path)
        os.environ["KONFAI_STATE"] = str(State.PREDICTION)
        

        def save():
            if os.path.exists("Predictions"):
                shutil.copytree("Predictions", args.output, dirs_exist_ok=True)
        from konfai.predictor import Predictor
        classname = Predictor
    elif args.command == "eval":

        for idx, file in enumerate(list_supported_files(args.gt)):
            dataset.write("Reference", "P{:03d}".format(idx), sitk.ReadImage(str(file)))
        if not args.mask:
            names = dataset.get_names("Volume")
            for name in names:
                data, attr = dataset.read_data("Volume", name)
                dataset.write("Mask", name, np.ones_like(data), attr)
        else:
            for idx, file in enumerate(list_supported_files(args.mask)):
                dataset.write("Mask", "P{:03d}".format(idx), sitk.ReadImage(str(file)))

        model_hf.install_evaluation()
        os.environ["KONFAI_ROOT"] = "Evaluator"

        os.environ["KONFAI_config_file"] = "Evaluation.yml"
        os.environ["KONFAI_STATE"] = str(State.EVALUATION)
        from konfai.evaluator import Evaluator
        def save():
            if os.path.exists("./Evaluations"):
                shutil.copytree("./Evaluations", args.output, dirs_exist_ok=True)
        
        classname = Evaluator
        
    elif args.command == "uncertainty":
        model_hf.install_uncertainty()
        os.environ["KONFAI_ROOT"] = "Evaluator"

        os.environ["KONFAI_config_file"] = "Uncertainty.yml"
        os.environ["KONFAI_STATE"] = str(State.EVALUATION)
        from konfai.evaluator import Evaluator

        def save():
            if os.path.exists("./Evaluations"):
                shutil.copytree("./Evaluations", args.output, dirs_exist_ok=True)

        classname = Evaluator
    elif args.command == "fine-tune":
        models_path = model_hf.install_fine_tune()
        os.environ["KONFAI_ROOT"] = "Trainer"

        os.environ["KONFAI_config_file"] = "FineTuning.yml"
        os.environ["KONFAI_MODEL"] = ":".join(models_path)
        os.environ["KONFAI_STATE"] = str(State.FINE_TUNING)
        def save():
            pass
        from konfai.trainer import Trainer
        classname = Trainer
    else:
        raise ValueError(f"Unknown command: {args.command}")


    return partial(classname, config=config_file()), args.tmp_dir, save


def setup(parser: argparse.ArgumentParser) -> DistributedObject:
    # KONFAI arguments
    konfai = parser.add_argument_group("KonfAI arguments")
    konfai.add_argument("type", type=State, choices=list(State))
    konfai.add_argument("-y", action="store_true", help="Accept overwrite")
    konfai.add_argument("-tb", action="store_true", help="Start TensorBoard")
    konfai.add_argument("-c", "--config", type=str, default="None", help="Configuration file location")
    konfai.add_argument(
        "-g",
        "--gpu",
        type=str,
        default=(os.environ["CUDA_VISIBLE_DEVICES"] if "CUDA_VISIBLE_DEVICES" in os.environ else ""),
        help="List of GPU",
    )
    konfai.add_argument("-cpu", "--cpu", type=str, default="1", help="Number of cores")
    konfai.add_argument(
        "--num-workers",
        "--num_workers",
        default=0,
        type=int,
        help="Number of workers per DataLoader & GPU",
    )
    konfai.add_argument(
        "-models_dir",
        "--MODELS_DIRECTORY",
        type=str,
        default="./Models/",
        help="Models location",
    )
    konfai.add_argument(
        "-checkpoints_dir",
        "--CHECKPOINTS_DIRECTORY",
        type=str,
        default="./Checkpoints/",
        help="Checkpoints location",
    )
    konfai.add_argument("-model", "--MODEL", type=str, default="", help="URL Model")
    konfai.add_argument(
        "-predictions_dir",
        "--PREDICTIONS_DIRECTORY",
        type=str,
        default="./Predictions/",
        help="Predictions location",
    )
    konfai.add_argument(
        "-evaluation_dir",
        "--EVALUATIONS_DIRECTORY",
        type=str,
        default="./Evaluations/",
        help="Evaluations location",
    )
    konfai.add_argument(
        "-statistics_dir",
        "--STATISTICS_DIRECTORY",
        type=str,
        default="./Statistics/",
        help="Statistics location",
    )
    konfai.add_argument(
        "-setups_dir",
        "--SETUPS_DIRECTORY",
        type=str,
        default="./Setups/",
        help="Setups location",
    )

    konfai.add_argument("-log", action="store_true", help="Enable logging to a file")
    konfai.add_argument("-quiet", action="store_false", help="Suppress console output for a quieter execution")

    konfai.add_argument("--version", action="version", version=importlib.metadata.version("konfai"))
    args = parser.parse_args()
    config = vars(args)

    os.environ["CUDA_VISIBLE_DEVICES"] = config["gpu"]
    os.environ["KONFAI_NB_CORES"] = config["cpu"]

    os.environ["KONFAI_WORKERS"] = str(config["num_workers"])
    os.environ["KONFAI_MODELS_DIRECTORY"] = config["MODELS_DIRECTORY"]
    os.environ["KONFAI_CHECKPOINTS_DIRECTORY"] = config["CHECKPOINTS_DIRECTORY"]
    os.environ["KONFAI_PREDICTIONS_DIRECTORY"] = config["PREDICTIONS_DIRECTORY"]
    os.environ["KONFAI_EVALUATIONS_DIRECTORY"] = config["EVALUATIONS_DIRECTORY"]
    os.environ["KONFAI_STATISTICS_DIRECTORY"] = config["STATISTICS_DIRECTORY"]
    os.environ["KONFAI_SETUPS_DIRECTORY"] = config["SETUPS_DIRECTORY"]

    os.environ["KONFAI_STATE"] = str(config["type"])

    os.environ["KONFAI_MODEL"] = config["MODEL"]

    os.environ["KONFAI_OVERWRITE"] = str(config["y"])
    os.environ["KONFAI_CONFIG_MODE"] = "Done"
    if config["tb"]:
        os.environ["KONFAI_TENSORBOARD_PORT"] = str(find_free_port())

    os.environ["KONFAI_VERBOSE"] = str(config["quiet"])

    if config["config"] == "None":
        if config["type"] is State.PREDICTION:
            os.environ["KONFAI_config_file"] = "Prediction.yml"
        elif config["type"] is State.EVALUATION:
            os.environ["KONFAI_config_file"] = "Evaluation.yml"
        else:
            os.environ["KONFAI_config_file"] = "Config.yml"
    else:
        os.environ["KONFAI_config_file"] = config["config"]
    torch.autograd.set_detect_anomaly(True)
    os.environ["CUDA_LAUNCH_BLOCKING"] = "1"

    if config["type"] is State.PREDICTION:
        from konfai.predictor import Predictor

        os.environ["KONFAI_ROOT"] = "Predictor"
        return Predictor(config=config_file())
    elif config["type"] is State.EVALUATION:
        from konfai.evaluator import Evaluator

        os.environ["KONFAI_ROOT"] = "Evaluator"
        return Evaluator(config=config_file())
    else:
        from konfai.trainer import Trainer

        os.environ["KONFAI_ROOT"] = "Trainer"
        return Trainer(config=config_file())


def setup_gpu(world_size: int, port: int, rank: int | None = None) -> tuple[int | None, int | None]:
    try:
        nodelist = os.getenv("SLURM_JOB_NODELIST")
        if nodelist is None:
            raise RuntimeError("SLURM_JOB_NODELIST is not set.")
        scontrol_path = shutil.which("scontrol")
        if scontrol_path is None:
            raise FileNotFoundError("scontrol not found in PATH")
        host_name = subprocess.check_output(
            [scontrol_path, "show", "hostnames", nodelist], text=True, stderr=subprocess.DEVNULL
        ).strip()  # nosec B603
    except Exception:
        host_name = "localhost"
    if rank is None:
        import submitit

        job_env = submitit.JobEnvironment()
        global_rank = job_env.global_rank
        local_rank = job_env.local_rank
    else:
        global_rank = rank
        local_rank = rank
    if global_rank >= world_size:
        return None, None
    # print("tcp://{}:{}".format(host_name, port))
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        dist.init_process_group(
            "nccl",
            rank=global_rank,
            init_method=f"tcp://{host_name}:{port}",
            world_size=world_size,
        )
    else:
        if not dist.is_initialized():
            dist.init_process_group(
                backend="gloo",
                init_method=f"tcp://{host_name}:{port}",
                rank=global_rank,
                world_size=world_size,
            )
    return global_rank, local_rank


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def cleanup():
    if dist.is_initialized():
        dist.destroy_process_group()


def synchronize_data(world_size: int, gpu: int, data: Any) -> list[Any]:
    if torch.cuda.is_available():
        outputs: list[dict[str, tuple[dict[str, float], dict[str, float]]] | None] = [None] * world_size
        torch.cuda.set_device(gpu)
        dist.all_gather_object(outputs, data)
    else:
        outputs = [data]
    return outputs


def _resample(data: torch.Tensor, size: list[int]) -> torch.Tensor:
    if data.dtype == torch.uint8:
        mode = "nearest"
    elif len(data.shape) < 4:
        mode = "bilinear"
    else:
        mode = "trilinear"
    return (
        F.interpolate(
            data.type(torch.float32).unsqueeze(0),
            size=tuple(reversed(size)),
            mode=mode,
        )
        .squeeze(0)
        .type(data.dtype)
    )


def _affine_matrix(matrix: torch.Tensor, translation: torch.Tensor) -> torch.Tensor:
    return torch.cat(
        (
            torch.cat((matrix, translation.unsqueeze(0).T), dim=1),
            torch.tensor([[0, 0, 0, 1]]),
        ),
        dim=0,
    )


def _resample_affine(data: torch.Tensor, matrix: torch.Tensor):
    if data.dtype == torch.uint8:
        mode = "nearest"
    else:
        mode = "bilinear"
    return (
        F.grid_sample(
            data.unsqueeze(0).type(torch.float32),
            F.affine_grid(
                matrix[:, :-1, ...].type(torch.float32),
                [1] + list(data.shape),
                align_corners=True,
            ),
            align_corners=True,
            mode=mode,
            padding_mode="reflection",
        )
        .squeeze(0)
        .type(data.dtype)
    )


def download_url(model_name: str, url: str) -> str:
    spec = importlib.util.find_spec("konfai")
    if spec is None or spec.submodule_search_locations is None:
        raise ImportError("Could not locate 'konfai' package")
    locations = spec.submodule_search_locations
    if not isinstance(locations, list) or not locations:
        raise ImportError("No valid submodule_search_locations found")
    base_path = Path(locations[0]) / "metric" / "models"
    os.makedirs(base_path, exist_ok=True)

    subdirs = Path(model_name).parent
    model_dir = base_path / subdirs
    model_dir.mkdir(exist_ok=True)
    filetmp = model_dir / ("tmp_" + str(Path(model_name).name))
    file = model_dir / Path(model_name).name
    if file.exists():
        return str(file)

    try:
        print(f"[FOCUS] Downloading {model_name} to {file}")
        with requests.get(url + model_name, stream=True, timeout=10) as r:
            r.raise_for_status()
            total = int(r.headers.get("content-length", 0))
            with open(filetmp, "wb") as f:
                with tqdm(
                    total=total,
                    unit="B",
                    unit_scale=True,
                    desc=f"Downloading {model_name}",
                ) as pbar:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)
                        pbar.update(len(chunk))
        shutil.copy2(filetmp, file)
        print("Download finished.")
    except Exception as e:
        raise e
    finally:
        if filetmp.exists():
            os.remove(filetmp)
    return str(file)


SUPPORTED_EXTENSIONS = [
    "mha",
    "mhd",  # MetaImage
    "nii",
    "nii.gz",  # NIfTI
    "nrrd",
    "nrrd.gz",  # NRRD
    "gipl",
    "gipl.gz",  # GIPL
    "hdr",
    "img",  # Analyze
    "dcm",  # DICOM (si GDCM activé)
    "tif",
    "tiff",  # TIFF
    "png",
    "jpg",
    "jpeg",
    "bmp",  # 2D formats
    "h5",
    "itk.txt",
    ".fcsv",
    ".xml",
    ".vtk",
    ".npy",
]


class KonfAIError(Exception):

    def __init__(self, type_error: str, *messages: str) -> None:
        super().__init__(
            f"\n[{type_error}] {messages[0]}" + ("\n" if len(messages) > 0 else "") + "\n→\t".join(messages[1:])
        )


class ConfigError(KonfAIError):

    def __init__(self, *message) -> None:
        super().__init__("Config", *message)


class DatasetManagerError(KonfAIError):

    def __init__(self, *message) -> None:
        super().__init__("DatasetManager", *message)


class MeasureError(KonfAIError):

    def __init__(self, *message) -> None:
        super().__init__("Measure", *message)


class TrainerError(KonfAIError):

    def __init__(self, *message) -> None:
        super().__init__("Trainer", *message)


class AugmentationError(KonfAIError):

    def __init__(self, *message) -> None:
        super().__init__("Augmentation", *message)


class EvaluatorError(KonfAIError):

    def __init__(self, *message) -> None:
        super().__init__("Evaluator", *message)


class PredictorError(KonfAIError):

    def __init__(self, *message) -> None:
        super().__init__("Predictor", *message)


class TransformError(KonfAIError):

    def __init__(self, *message) -> None:
        super().__init__("Transform", *message)


class RepositoryHFError(KonfAIError):

    def __init__(self, *message) -> None:
        super().__init__("HF", *message)


def get_available_models_on_hf_repo(repo_id: str) -> list[str]:
    api = HfApi()
    model_names = []
    try:
        tree = api.list_repo_tree(repo_id=repo_id)
    except Exception as e:
        raise RepositoryHFError(f"Unable to access repository '{repo_id}': {e}")
    for entry in tree:
        model_name = entry.path
        if isinstance(entry, RepoFolder) and is_model_repo(repo_id, model_name)[0]:
            model_names.append(model_name)
    return model_names


def is_model_repo(repo_id: str, model_name: str) -> tuple[bool, str, int]:
    """
    Check whether the Hugging Face repository structure is valid for KonfAI.
    Required files:
    - at least one .pt file in the model folder
    - a Prediction.yml file
    - a metadata.json file
    """
    api = HfApi()
    fold_names = []
    found_prediction_file = False
    found_metadata_file = False

    try:
        tree = api.list_repo_tree(repo_id=repo_id, path_in_repo=model_name)
    except Exception as e:
        return False, f"Unable to access repository '{repo_id}': {e}", 0

    for file in tree:
        if file.path.endswith(".pt"):
            fold_names.append(file.path)
        elif file.path.endswith("Prediction.yml"):
            found_prediction_file = True
        elif file.path.endswith("metadata.json"):
            found_metadata_file = True

    if len(fold_names) == 0:
        return False, f"No '.pt' model files were found in '{repo_id}/{model_name}'.", 0

    if not found_prediction_file:
        return False, f"Missing 'Prediction.yml' in '{repo_id}/{model_name}'.", 0

    if not found_metadata_file:
        return False, f"Missing 'metadata.json' in '{repo_id}/{model_name}'.", 0
    return True, "", len(fold_names)


class ModelHF:

    def __init__(self, model_name: str):
        if model_name and len(model_name.split(":")) != 2:
            raise RepositoryHFError(
                f"Invalid model name format in --config: '{model_name}'. "
                "Expected format is 'REPO_ID:NAME', e.g. 'VBoussot/ImpactSynth:MR'."
            )
        self.repo_id, self.model_name = model_name.split(":")
        _, err_message, self._number_of_models = is_model_repo(self.repo_id, self.model_name)
        if err_message:
            raise RepositoryHFError(err_message)
        model_metadata = self._read_metadata()

        required_keys = ["description", "short_description", "tta", "mc_dropout", "display_name"]
        missing = [k for k in required_keys if k not in model_metadata]
        if missing:
            raise RepositoryHFError(f"Missing keys in metadata.json: {', '.join(missing)}")

        self._description = str(model_metadata["description"])
        self._short_description = str(model_metadata["short_description"])

        try:
            self._maximum_tta = int(model_metadata["tta"])
        except Exception:
            raise RepositoryHFError("The field 'tta' must be an integer.")

        try:
            self._mc_dropout = int(model_metadata["mc_dropout"])
        except Exception:
            raise RepositoryHFError("The field 'mc_dropout' must be an integer.")

        self._display_name = str(model_metadata["display_name"])

    def has_capabilities(self) -> tuple[bool, bool]:
        api = HfApi()
        tree = api.list_repo_tree(repo_id=self.repo_id, path_in_repo=self.model_name)
        evaluation_support = False
        uncertainty_support = False

        for file in tree:
            if file.path.endswith("Evaluation.yml"):
                evaluation_support = True
            elif file.path.endswith("Uncertainty.yml"):
                uncertainty_support = True
        return evaluation_support, uncertainty_support

    def _read_metadata(self) -> dict[str, str]:
        metadata_file_path = hf_hub_download(
            repo_id=self.repo_id, filename=f"{self.model_name}/metadata.json", repo_type="model", revision=None
        )  # nosec B615
        with open(metadata_file_path, encoding="utf-8") as f:
            model_metadata = json.load(f)
        return model_metadata

    def set_number_of_augmentation(self, inference_file_path: str, new_value: int) -> None:
        new_value = int(np.clip(new_value, 0, self._maximum_tta))
        yaml = YAML()
        with open(inference_file_path) as f:
            data = yaml.load(f)

        if new_value > 0:
            tmp = data["Predictor"]["Dataset"]["augmentations"]
            if "DataAugmentation_0" in tmp:
                tmp["DataAugmentation_0"]["nb"] = new_value

        else:
            data["Predictor"]["Dataset"]["augmentations"] = {}
        with open(inference_file_path, "w") as f:
            yaml.dump(data, f)

    def download_inference(self, number_of_model: int) -> tuple[list[str], str, list[str]]:
        api = HfApi()
        models_path = []
        codes_path = []
        i = 0
        for filename in api.list_repo_tree(repo_id=self.repo_id, path_in_repo=self.model_name):
            if filename.path.endswith(".pt"):
                i += 1
                if i > number_of_model:
                    continue

            file_path = hf_hub_download(
                repo_id=self.repo_id, filename=filename.path, repo_type="model", revision=None
            )  # nosec B615
            if "Prediction.yml" in filename.path:
                inference_file_path = file_path
            elif filename.path.endswith(".pt"):
                models_path.append(file_path)
            elif "requirements.txt" in filename.path:
                with open(file_path, encoding="utf-8") as f:
                    required_lines = [line.strip() for line in f if line.strip() and not line.startswith("#")]

                    installed = {
                        dist.metadata["Name"].lower(): dist.version
                        for dist in importlib.metadata.distributions()
                        if dist.metadata.get("Name")
                    }

                    for line in required_lines:
                        req = Requirement(line)
                        name = req.name.lower()
                        installed_version_str = installed.get(name)
                        missing_or_outdated = []
                        if installed_version_str is None:
                            missing_or_outdated.append(line)
                            continue

                        if req.specifier:
                            if not req.specifier.contains(installed_version_str, prereleases=True):
                                missing_or_outdated.append(line)

                    if missing_or_outdated:
                        try:
                            subprocess.check_call(
                                [sys.executable, "-m", "pip", "install", *missing_or_outdated],  # nosec B603
                            )
                        except subprocess.CalledProcessError as e:
                            raise RepositoryHFError(f"Failed to install packages: {e}") from e
            elif "metadata.json" not in filename.path:
                codes_path.append(file_path)
        return models_path, inference_file_path, codes_path

    def download_evaluation(self) -> tuple[str, list[str]]:
        api = HfApi()
        codes_path = []
        for filename in api.list_repo_tree(repo_id=self.repo_id, path_in_repo=self.model_name):
            if "Evaluation.yml" in filename.path:
                evaluation_file_path = hf_hub_download(
                    repo_id=self.repo_id, filename=filename.path, repo_type="model", revision=None
                )  # nosec B615
            elif filename.path.endswith(".py"):
                file_path = hf_hub_download(
                    repo_id=self.repo_id, filename=filename.path, repo_type="model", revision=None
                )  # nosec B615
                codes_path.append(file_path)
        return evaluation_file_path, codes_path

    def download_uncertainty(self) -> tuple[str, list[str]]:
        api = HfApi()
        codes_path = []
        for filename in api.list_repo_tree(repo_id=self.repo_id, path_in_repo=self.model_name):
            if "Uncertainty.yml" in filename.path:
                uncertainty_file_path = hf_hub_download(
                    repo_id=self.repo_id, filename=filename.path, repo_type="model", revision=None
                )  # nosec B615
            elif filename.path.endswith(".py"):
                file_path = hf_hub_download(
                    repo_id=self.repo_id, filename=filename.path, repo_type="model", revision=None
                )  # nosec B615
                codes_path.append(file_path)
        return uncertainty_file_path, codes_path

    def install_inference(
        self, number_of_augmentation: int, number_of_model: int, number_of_mc_dropout: int
    ) -> list[str]:
        if not number_of_model:
            number_of_model = self._number_of_models
        else:
            try:
                number_of_model = int(number_of_model)
            except ValueError:
                raise RepositoryHFError(
                    f"Invalid value provided for '--MODEL': '{number_of_model}'. ",
                    "The value must be an integer (e.g. 1, 2, 3).",
                )
        models_path, inference_file_path, codes_path = self.download_inference(number_of_model)

        shutil.copy2(inference_file_path, "./Prediction.yml")
        self.set_number_of_augmentation("./Prediction.yml", number_of_augmentation)
        for code_path in codes_path:
            if code_path.endswith(".py"):
                shutil.copy2(code_path, "./{}".format(code_path.split("/")[-1]))

        return models_path

    def install_evaluation(self) -> None:
        evaluation_file_path, codes_path = self.download_evaluation()
        shutil.copy2(evaluation_file_path, "./Evaluation.yml")
        for code_path in codes_path:
            if code_path.endswith(".py"):
                shutil.copy2(code_path, "./{}".format(code_path.split("/")[-1]))

    def install_uncertainty(self) -> None:
        uncertainty_file_path, codes_path = self.download_uncertainty()
        shutil.copy2(uncertainty_file_path, "./Uncertainty.yml")
        for code_path in codes_path:
            if code_path.endswith(".py"):
                shutil.copy2(code_path, "./{}".format(code_path.split("/")[-1]))

    def install_fine_tune(self):
        pass

    def get_display_name(self):
        return self._display_name

    def get_maximum_tta(self):
        return self._maximum_tta

    def get_mc_dropout(self):
        return self._mc_dropout

    def get_number_of_models(self):
        return self._number_of_models

    def get_description(self):
        return self._description

    def get_short_description(self):
        return self._short_description



from functools import partial
from typing import cast

import numpy as np
import torch

from konfai.data import augmentation
from konfai.data.patching import Attribute, ModelPatch
from konfai.models.generation.ddpm import DDPM
from konfai.models.segmentation import NestedUNet, UNet
from konfai.network import blocks, network


class Discriminator(network.Network):

    class DiscriminatorNLayers(network.ModuleArgsDict):

        def __init__(self, channels: list[int], strides: list[int], dim: int) -> None:
            super().__init__()
            block_config = partial(
                blocks.BlockConfig,
                kernel_size=4,
                padding=1,
                bias=False,
                activation=partial(torch.nn.LeakyReLU, negative_slope=0.2, inplace=True),
                norm_mode=blocks.NormMode.SYNCBATCH,
            )
            for i, (in_channels, out_channels, stride) in enumerate(zip(channels, channels[1:], strides)):
                self.add_module(
                    f"Layer_{i}",
                    blocks.ConvBlock(in_channels, out_channels, [block_config(stride=stride)], dim),
                )

    class DiscriminatorHead(network.ModuleArgsDict):

        def __init__(self, channels: int, dim: int) -> None:
            super().__init__()
            self.add_module(
                "Conv",
                blocks.get_torch_module("Conv", dim)(
                    in_channels=channels,
                    out_channels=1,
                    kernel_size=4,
                    stride=1,
                    padding=1,
                ),
            )
            # self.add_module("AdaptiveAvgPool", blocks.get_torch_module("AdaptiveAvgPool", dim)(tuple([1]*dim)))
            # self.add_module("Flatten", torch.nn.Flatten(1))

    class DiscriminatorBlock(network.ModuleArgsDict):

        def __init__(
            self,
            channels: list[int] = [1, 16, 32, 64, 64],
            strides: list[int] = [2, 2, 2, 1],
            dim: int = 3,
        ) -> None:
            super().__init__()
            self.add_module("Layers", Discriminator.DiscriminatorNLayers(channels, strides, dim))
            self.add_module("Head", Discriminator.DiscriminatorHead(channels[-1], dim))

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        channels: list[int] = [1, 16, 32, 64, 64],
        strides: list[int] = [2, 2, 2, 1],
        nb_batch_per_step: int = 1,
        dim: int = 3,
    ) -> None:
        super().__init__(
            in_channels=1,
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            nb_batch_per_step=nb_batch_per_step,
            dim=dim,
            init_type="kaiming",
        )
        self.add_module(
            "DiscriminatorModel",
            Discriminator.DiscriminatorBlock(channels, strides, dim),
        )


class DiscriminatorADA(network.Network):

    class DDPMTE(torch.nn.Module):

        def __init__(self, in_channels: int, out_channels: int) -> None:
            super().__init__()
            self.linear_0 = torch.nn.Linear(in_channels, out_channels)
            self.siLU = torch.nn.SiLU()
            self.linear_1 = torch.nn.Linear(out_channels, out_channels)

        def forward(self, tensor: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
            return tensor + self.linear_1(self.siLU(self.linear_0(t))).reshape(
                tensor.shape[0], -1, *[1 for _ in range(len(tensor.shape) - 2)]
            )

    class DiscriminatorNLayers(network.ModuleArgsDict):

        def __init__(
            self,
            channels: list[int],
            strides: list[int],
            time_embedding_dim: int,
            dim: int,
        ) -> None:
            super().__init__()
            block_config = partial(
                blocks.BlockConfig,
                kernel_size=4,
                padding=1,
                bias=False,
                activation=partial(torch.nn.LeakyReLU, negative_slope=0.2, inplace=True),
                norm_mode=blocks.NormMode.SYNCBATCH,
            )
            for i, (in_channels, out_channels, stride) in enumerate(zip(channels, channels[1:], strides)):
                self.add_module(
                    f"Te_{i}",
                    DiscriminatorADA.DDPMTE(time_embedding_dim, in_channels),
                    in_branch=[0, 1],
                )
                self.add_module(
                    f"Layer_{i}",
                    blocks.ConvBlock(in_channels, out_channels, [block_config(stride=stride)], dim),
                )

    class DiscriminatorHead(network.ModuleArgsDict):

        def __init__(self, channels: int, dim: int) -> None:
            super().__init__()
            self.add_module(
                "Conv",
                blocks.get_torch_module("Conv", dim)(
                    in_channels=channels,
                    out_channels=1,
                    kernel_size=4,
                    stride=1,
                    padding=1,
                ),
            )
            # self.add_module("AdaptiveAvgPool", blocks.get_torch_module("AdaptiveAvgPool", dim)(tuple([1]*dim)))
            # self.add_module("Flatten", torch.nn.Flatten(1))

    class UpdateP(torch.nn.Module):

        def __init__(self):
            super().__init__()
            self._it = 0
            self.n = 4
            self.ada_target = 0.25
            self.ada_interval = 0.001
            self.ada_kimg = 500

            self.measure = None
            self.names = []
            self.p = 0

        def set_measure(self, measure: network.Measure, names: list[str]):
            self.measure = measure
            self.names = names

        def forward(self, tensor: torch.Tensor) -> torch.Tensor:
            if self.measure is not None and self._it % self.n == 0:
                value = sum([v for k, v in self.measure.get_last_values(self.n).items() if k in self.names])
                adjust = np.sign(self.ada_target - value) * (self.ada_interval)
                self.p += adjust
                self.p = np.clip(self.p, 0, 1)
            self._it += 1
            return torch.tensor(self.p).to(tensor.device)

    class DiscriminatorAugmentation(torch.nn.Module):

        def __init__(self, dim: int):
            super().__init__()

            self.data_augmentations: dict[augmentation.DataAugmentation, float] = {}
            pixel_blitting = {
                augmentation.Flip([1 / 3] * 3 if dim == 3 else [1 / 2] * 2): 0,
                augmentation.Rotate(a_min=0, a_max=360, is_quarter=True): 0,
                augmentation.Translate(-5, 5, is_int=True): 0,
            }

            self.data_augmentations.update(
                {cast(augmentation.DataAugmentation, k): v for k, v in pixel_blitting.items()}
            )
            geometric = {
                augmentation.Scale(0.2): 0,
                augmentation.Rotate(a_min=0, a_max=360): 0,
                augmentation.Scale(0.2): 0,
                augmentation.Rotate(a_min=0, a_max=360): 0,
                augmentation.Translate(-5, 5): 0,
                augmentation.Elastix(16, 16): 0.5,
            }
            self.data_augmentations.update({cast(augmentation.DataAugmentation, k): v for k, v in geometric.items()})
            color = {
                augmentation.Brightness(0.2): 0.0,
                augmentation.Contrast(0.5): 0.0,
                augmentation.Saturation(1): 0.0,
                augmentation.HUE(1): 0.0,
                augmentation.LumaFlip(): 0.0,
            }
            self.data_augmentations.update({cast(augmentation.DataAugmentation, k): v for k, v in color.items()})

            corruptions = {
                augmentation.Noise(1): 1,
                augmentation.CutOUT(0.5, 1, -1): 0.3,
            }
            self.data_augmentations.update({cast(augmentation.DataAugmentation, k): v for k, v in corruptions.items()})

        def _set_p(self, prob: float):
            for aug, p in self.data_augmentations.items():
                aug.load(prob * p)

        def forward(self, tensor: torch.Tensor, prob: torch.Tensor) -> torch.Tensor:
            self._set_p(prob.item())
            out = tensor
            for aug in self.data_augmentations.keys():
                aug.state_init(
                    None,
                    [tensor.shape[2:]] * tensor.shape[0],
                    [Attribute()] * tensor.shape[0],
                )
                out = aug("", 0, list(out))
            return torch.cat([data.unsqueeze(0) for data in out], 0)

    class DiscriminatorBlock(network.ModuleArgsDict):

        def __init__(
            self,
            channels: list[int] = [1, 16, 32, 64, 64],
            strides: list[int] = [2, 2, 2, 1],
            dim: int = 3,
        ) -> None:
            super().__init__()
            self.add_module("Prob", DiscriminatorADA.UpdateP(), out_branch=["p"])
            self.add_module(
                "Sample",
                DiscriminatorADA.DiscriminatorAugmentation(dim),
                in_branch=[0, "p"],
            )
            self.add_module(
                "t",
                DDPM.DDPMTimeEmbedding(1000, 100),
                in_branch=[0, "p"],
                out_branch=["te"],
            )
            self.add_module(
                "Layers",
                DiscriminatorADA.DiscriminatorNLayers(channels, strides, 100, dim),
                in_branch=[0, "te"],
            )
            self.add_module("Head", DiscriminatorADA.DiscriminatorHead(channels[-1], dim))

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        channels: list[int] = [1, 16, 32, 64, 64],
        strides: list[int] = [2, 2, 2, 1],
        nb_batch_per_step: int = 1,
        dim: int = 3,
    ) -> None:
        super().__init__(
            in_channels=1,
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            nb_batch_per_step=nb_batch_per_step,
            dim=dim,
            init_type="kaiming",
        )
        self.add_module(
            "DiscriminatorModel",
            DiscriminatorADA.DiscriminatorBlock(channels, strides, dim),
        )

    def initialized(self):
        self["DiscriminatorModel"]["Prob"].set_measure(
            self.measure,
            ["Discriminator_B.DiscriminatorModel.Head.Conv:None:PatchGanLoss"],
        )


class GeneratorV1(network.Network):

    class GeneratorStem(network.ModuleArgsDict):

        def __init__(self, in_channels: int, out_channels: int, dim: int) -> None:
            super().__init__()
            self.add_module(
                "ConvBlock",
                blocks.ConvBlock(
                    in_channels,
                    out_channels,
                    block_configs=[blocks.BlockConfig(bias=False, activation="ReLU", norm_mode="SYNCBATCH")],
                    dim=dim,
                ),
            )

    class GeneratorHead(network.ModuleArgsDict):

        def __init__(self, in_channels: int, out_channels: int, dim: int) -> None:
            super().__init__()
            self.add_module(
                "ConvBlock",
                blocks.ConvBlock(
                    in_channels,
                    in_channels,
                    block_configs=[blocks.BlockConfig(bias=False, activation="ReLU", norm_mode="SYNCBATCH")],
                    dim=dim,
                ),
            )
            self.add_module(
                "Conv",
                blocks.get_torch_module("Conv", dim)(in_channels, out_channels, kernel_size=1, bias=False),
            )
            self.add_module("Tanh", torch.nn.Tanh())

    class GeneratorDownSample(network.ModuleArgsDict):

        def __init__(self, in_channels: int, out_channels: int, dim: int) -> None:
            super().__init__()
            self.add_module(
                "ConvBlock",
                blocks.ConvBlock(
                    in_channels,
                    out_channels,
                    block_configs=[
                        blocks.BlockConfig(
                            stride=2,
                            bias=False,
                            activation="ReLU",
                            norm_mode="SYNCBATCH",
                        )
                    ],
                    dim=dim,
                ),
            )

    class GeneratorUpSample(network.ModuleArgsDict):

        def __init__(self, in_channels: int, out_channels: int, dim: int) -> None:
            super().__init__()
            self.add_module(
                "ConvBlock",
                blocks.ConvBlock(
                    in_channels,
                    out_channels,
                    block_configs=[blocks.BlockConfig(bias=False, activation="ReLU", norm_mode="SYNCBATCH")],
                    dim=dim,
                ),
            )
            self.add_module(
                "Upsample",
                torch.nn.Upsample(scale_factor=2, mode="bilinear" if dim < 3 else "trilinear"),
            )

    class GeneratorEncoder(network.ModuleArgsDict):
        def __init__(self, channels: list[int], dim: int) -> None:
            super().__init__()
            for i, (in_channels, out_channels) in enumerate(zip(channels, channels[1:])):
                self.add_module(
                    f"DownSample_{i}",
                    GeneratorV1.GeneratorDownSample(in_channels=in_channels, out_channels=out_channels, dim=dim),
                )

    class GeneratorResnetBlock(network.ModuleArgsDict):

        def __init__(self, channels: int, dim: int):
            super().__init__()
            self.add_module(
                "Conv_0",
                blocks.get_torch_module("Conv", dim)(channels, channels, kernel_size=3, padding=1, bias=False),
            )
            self.add_module("Norm_0", torch.nn.SyncBatchNorm(channels))
            self.add_module("Activation_0", torch.nn.LeakyReLU(0.2, inplace=True))
            # self.add_module("Norm", torch.nn.LeakyReLU(0.2, inplace=True))

            self.add_module(
                "Conv_1",
                blocks.get_torch_module("Conv", dim)(channels, channels, kernel_size=3, padding=1, bias=False),
            )
            self.add_module("Norm_1", torch.nn.SyncBatchNorm(channels))
            self.add_module("Residual", blocks.Add(), in_branch=[0, 1])

    class GeneratorNResnetBlock(network.ModuleArgsDict):

        def __init__(self, channels: int, nb_conv: int, dim: int) -> None:
            super().__init__()
            for i in range(nb_conv):
                self.add_module(
                    f"ResnetBlock_{i}",
                    GeneratorV1.GeneratorResnetBlock(channels=channels, dim=dim),
                )

    class GeneratorDecoder(network.ModuleArgsDict):
        def __init__(self, channels: list[int], dim: int) -> None:
            super().__init__()
            for i, (in_channels, out_channels) in enumerate(zip(reversed(channels), reversed(channels[:-1]))):
                self.add_module(
                    f"UpSample_{i}",
                    GeneratorV1.GeneratorUpSample(in_channels=in_channels, out_channels=out_channels, dim=dim),
                )

    class GeneratorAutoEncoder(network.ModuleArgsDict):

        def __init__(self, ngf: int, dim: int) -> None:
            super().__init__()
            channels = [ngf, ngf * 2]
            self.add_module("Encoder", GeneratorV1.GeneratorEncoder(channels, dim))
            self.add_module(
                "NResBlock",
                GeneratorV1.GeneratorNResnetBlock(channels=channels[-1], nb_conv=6, dim=dim),
            )
            self.add_module("Decoder", GeneratorV1.GeneratorDecoder(channels, dim))

    class GeneratorBlock(network.ModuleArgsDict):

        def __init__(self, ngf: int, dim: int) -> None:
            super().__init__()
            self.add_module("Stem", GeneratorV1.GeneratorStem(3, ngf, dim))
            self.add_module("AutoEncoder", GeneratorV1.GeneratorAutoEncoder(ngf, dim))
            self.add_module(
                "Head",
                GeneratorV1.GeneratorHead(in_channels=ngf, out_channels=1, dim=dim),
            )

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        patch: ModelPatch = ModelPatch(),
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        dim: int = 3,
    ) -> None:
        super().__init__(
            optimizer=optimizer,
            in_channels=3,
            schedulers=schedulers,
            patch=patch,
            outputs_criterions=outputs_criterions,
            dim=dim,
        )
        self.add_module("GeneratorModel", GeneratorV1.GeneratorBlock(32, dim))


class GeneratorV2(network.Network):

    class NestedUNetHead(network.ModuleArgsDict):

        def __init__(self, in_channels: list[int], dim: int) -> None:
            super().__init__()
            self.add_module(
                "Conv",
                blocks.get_torch_module("Conv", dim)(
                    in_channels=in_channels[1],
                    out_channels=1,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                ),
            )
            self.add_module("Tanh", torch.nn.Tanh())

    class GeneratorBlock(network.ModuleArgsDict):

        def __init__(
            self,
            channels: list[int],
            block_config: blocks.BlockConfig,
            nb_conv_per_stage: int,
            downsample_mode: str,
            upsample_mode: str,
            attention: bool,
            block_type: str,
            dim: int,
        ) -> None:
            super().__init__()
            self.add_module(
                "UNetBlock_0",
                NestedUNet.NestedUNet.NestedUNetBlock(
                    channels,
                    nb_conv_per_stage,
                    block_config,
                    downsample_mode=blocks.DownsampleMode[downsample_mode],
                    upsample_mode=blocks.UpsampleMode[upsample_mode],
                    attention=attention,
                    block=blocks.ConvBlock if block_type == "Conv" else blocks.ResBlock,
                    dim=dim,
                ),
                out_branch=[f"X_0_{j + 1}" for j in range(len(channels) - 2)],
            )
            self.add_module(
                "Head",
                GeneratorV2.NestedUNetHead(channels[:2], dim=dim),
                in_branch=[f"X_0_{len(channels) - 2}"],
            )

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        patch: ModelPatch | None = None,
        channels: list[int] = [1, 64, 128, 256, 512, 1024],
        block_config: blocks.BlockConfig = blocks.BlockConfig(),
        nb_conv_per_stage: int = 2,
        downsample_mode: str = "MAXPOOL",
        upsample_mode: str = "CONV_TRANSPOSE",
        attention: bool = False,
        block_type: str = "Conv",
        dim: int = 3,
    ) -> None:
        super().__init__(
            in_channels=channels[0],
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            patch=patch,
            dim=dim,
        )
        self.add_module(
            "GeneratorModel",
            GeneratorV2.GeneratorBlock(
                channels,
                block_config,
                nb_conv_per_stage,
                downsample_mode,
                upsample_mode,
                attention,
                block_type,
                dim,
            ),
        )


class GeneratorV3(network.Network):

    class NestedUNetHead(network.ModuleArgsDict):

        def __init__(self, in_channels: list[int], dim: int) -> None:
            super().__init__()
            self.add_module(
                "Conv",
                blocks.get_torch_module("Conv", dim)(
                    in_channels=in_channels[1],
                    out_channels=1,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                ),
            )
            self.add_module("Tanh", torch.nn.Tanh())

    class GeneratorBlock(network.ModuleArgsDict):

        def __init__(
            self,
            channels: list[int],
            block_config: blocks.BlockConfig,
            nb_conv_per_stage: int,
            downsample_mode: str,
            upsample_mode: str,
            attention: bool,
            block_type: str,
            dim: int,
        ) -> None:
            super().__init__()
            self.add_module(
                "UNetBlock_0",
                UNet.UNetBlock(
                    channels,
                    nb_conv_per_stage,
                    block_config,
                    downsample_mode=blocks.DownsampleMode[downsample_mode],
                    upsample_mode=blocks.UpsampleMode[upsample_mode],
                    attention=attention,
                    block=blocks.ConvBlock if block_type == "Conv" else blocks.ResBlock,
                    nb_class=1,
                    dim=dim,
                ),
                out_branch=[f"X_0_{j + 1}" for j in range(len(channels) - 2)],
            )
            self.add_module(
                "Head",
                GeneratorV3.NestedUNetHead(channels[:2], dim=dim),
                in_branch=[f"X_0_{len(channels) - 2}"],
            )

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        patch: ModelPatch | None = None,
        channels: list[int] = [1, 64, 128, 256, 512, 1024],
        block_config: blocks.BlockConfig = blocks.BlockConfig(),
        nb_conv_per_stage: int = 2,
        downsample_mode: str = "MAXPOOL",
        upsample_mode: str = "CONV_TRANSPOSE",
        attention: bool = False,
        block_type: str = "Conv",
        dim: int = 3,
    ) -> None:
        super().__init__(
            in_channels=channels[0],
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            patch=patch,
            dim=dim,
        )
        self.add_module(
            "GeneratorModel",
            GeneratorV3.GeneratorBlock(
                channels,
                block_config,
                nb_conv_per_stage,
                downsample_mode,
                upsample_mode,
                attention,
                block_type,
                dim,
            ),
            out_branch=["pB"],
        )


class DiffusionGan(network.Network):

    def __init__(
        self,
        generator: GeneratorV1 = GeneratorV1(),
        discriminator: DiscriminatorADA = DiscriminatorADA(),
    ) -> None:
        super().__init__()
        self.add_module("Generator_A_to_B", generator, in_branch=[0], out_branch=["pB"])
        self.add_module(
            "Discriminator_B",
            discriminator,
            in_branch=[1],
            out_branch=[-1],
            requires_grad=True,
        )
        self.add_module("detach", blocks.Detach(), in_branch=["pB"], out_branch=["pB_detach"])
        self.add_module(
            "Discriminator_pB_detach",
            discriminator,
            in_branch=["pB_detach"],
            out_branch=[-1],
        )
        self.add_module(
            "Discriminator_pB",
            discriminator,
            in_branch=["pB"],
            out_branch=[-1],
            requires_grad=False,
        )


class DiffusionGanV2(network.Network):

    def __init__(
        self,
        generator: GeneratorV2 = GeneratorV2(),
        discriminator: Discriminator = Discriminator(),
    ) -> None:
        super().__init__()
        self.add_module("Generator_A_to_B", generator, in_branch=[0], out_branch=["pB"])
        self.add_module(
            "Discriminator_B",
            discriminator,
            in_branch=[1],
            out_branch=[-1],
            requires_grad=True,
        )
        self.add_module("detach", blocks.Detach(), in_branch=["pB"], out_branch=["pB_detach"])
        self.add_module(
            "Discriminator_pB_detach",
            discriminator,
            in_branch=["pB_detach"],
            out_branch=[-1],
        )
        self.add_module(
            "Discriminator_pB",
            discriminator,
            in_branch=["pB"],
            out_branch=[-1],
            requires_grad=False,
        )


class CycleGanDiscriminator(network.Network):

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        patch: ModelPatch | None = None,
        channels: list[int] = [1, 16, 32, 64, 64],
        strides: list[int] = [2, 2, 2, 1],
        dim: int = 3,
    ) -> None:
        super().__init__(
            in_channels=1,
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            patch=patch,
            dim=dim,
        )
        self.add_module(
            "Discriminator_A",
            Discriminator.DiscriminatorBlock(channels, strides, dim),
            in_branch=[0],
            out_branch=[0],
        )
        self.add_module(
            "Discriminator_B",
            Discriminator.DiscriminatorBlock(channels, strides, dim),
            in_branch=[1],
            out_branch=[1],
        )

    def initialized(self):
        self["Discriminator_A"]["Sample"].set_measure(
            self.measure,
            ["Discriminator.Discriminator_A.Head.Flatten:None:PatchGanLoss"],
        )
        self["Discriminator_B"]["Sample"].set_measure(
            self.measure,
            ["Discriminator.Discriminator_B.Head.Flatten:None:PatchGanLoss"],
        )


class CycleGanGeneratorV1(network.Network):

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        patch: ModelPatch | None = None,
        dim: int = 3,
    ) -> None:
        super().__init__(
            in_channels=1,
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            patch=patch,
            dim=dim,
        )
        self.add_module(
            "Generator_A_to_B",
            GeneratorV1.GeneratorBlock(32, dim),
            in_branch=[0],
            out_branch=["pB"],
        )
        self.add_module(
            "Generator_B_to_A",
            GeneratorV1.GeneratorBlock(32, dim),
            in_branch=[1],
            out_branch=["pA"],
        )


class CycleGanGeneratorV2(network.Network):

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        patch: ModelPatch | None = None,
        channels: list[int] = [1, 64, 128, 256, 512, 1024],
        block_config: blocks.BlockConfig = blocks.BlockConfig(),
        nb_conv_per_stage: int = 2,
        downsample_mode: str = "MAXPOOL",
        upsample_mode: str = "CONV_TRANSPOSE",
        attention: bool = False,
        block_type: str = "Conv",
        dim: int = 3,
    ) -> None:
        super().__init__(
            in_channels=1,
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            patch=patch,
            dim=dim,
        )
        self.add_module(
            "Generator_A_to_B",
            GeneratorV2.GeneratorBlock(
                channels,
                block_config,
                nb_conv_per_stage,
                downsample_mode,
                upsample_mode,
                attention,
                block_type,
                dim,
            ),
            in_branch=[0],
            out_branch=["pB"],
        )
        self.add_module(
            "Generator_B_to_A",
            GeneratorV2.GeneratorBlock(
                channels,
                block_config,
                nb_conv_per_stage,
                downsample_mode,
                upsample_mode,
                attention,
                block_type,
                dim,
            ),
            in_branch=[1],
            out_branch=["pA"],
        )


class CycleGanGeneratorV3(network.Network):

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        patch: ModelPatch | None = None,
        channels: list[int] = [1, 64, 128, 256, 512, 1024],
        block_config: blocks.BlockConfig = blocks.BlockConfig(),
        nb_conv_per_stage: int = 2,
        downsample_mode: str = "MAXPOOL",
        upsample_mode: str = "CONV_TRANSPOSE",
        attention: bool = False,
        block_type: str = "Conv",
        dim: int = 3,
    ) -> None:
        super().__init__(
            in_channels=1,
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            patch=patch,
            dim=dim,
        )
        self.add_module(
            "Generator_A_to_B",
            GeneratorV3.GeneratorBlock(
                channels,
                block_config,
                nb_conv_per_stage,
                downsample_mode,
                upsample_mode,
                attention,
                block_type,
                dim,
            ),
            in_branch=[0],
            out_branch=["pB"],
        )
        self.add_module(
            "Generator_B_to_A",
            GeneratorV3.GeneratorBlock(
                channels,
                block_config,
                nb_conv_per_stage,
                downsample_mode,
                upsample_mode,
                attention,
                block_type,
                dim,
            ),
            in_branch=[1],
            out_branch=["pA"],
        )


class DiffusionCycleGan(network.Network):

    def __init__(
        self,
        generators: CycleGanGeneratorV3 = CycleGanGeneratorV3(),
        discriminators: CycleGanDiscriminator = CycleGanDiscriminator(),
    ) -> None:
        super().__init__()
        self.add_module("Generator", generators, in_branch=[0, 1], out_branch=["pB", "pA"])
        self.add_module(
            "Discriminator",
            discriminators,
            in_branch=[0, 1],
            out_branch=[-1],
            requires_grad=True,
        )

        self.add_module("Generator_identity", generators, in_branch=[1, 0], out_branch=[-1])

        self.add_module("Generator_p", generators, in_branch=["pA", "pB"], out_branch=[-1])

        self.add_module("detach_pA", blocks.Detach(), in_branch=["pA"], out_branch=["pA_detach"])
        self.add_module("detach_pB", blocks.Detach(), in_branch=["pB"], out_branch=["pB_detach"])

        self.add_module(
            "Discriminator_p_detach",
            discriminators,
            in_branch=["pA_detach", "pB_detach"],
            out_branch=[-1],
        )
        self.add_module(
            "Discriminator_p",
            discriminators,
            in_branch=["pA", "pB"],
            out_branch=[-1],
            requires_grad=False,
        )



import torch
import torch.nn.functional as F  # noqa: N812
from torch.nn.parameter import Parameter

from konfai.models.segmentation import UNet
from konfai.network import blocks, network


class VoxelMorph(network.Network):

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        dim: int = 3,
        channels: list[int] = [4, 16, 32, 32, 32],
        block_config: blocks.BlockConfig = blocks.BlockConfig(),
        nb_conv_per_stage: int = 2,
        downsample_mode: str = "MAXPOOL",
        upsample_mode: str = "CONV_TRANSPOSE",
        attention: bool = False,
        shape: list[int] = [192, 192, 192],
        int_steps: int = 7,
        int_downsize: int = 2,
        nb_batch_per_step: int = 1,
        rigid: bool = False,
    ):
        super().__init__(
            in_channels=channels[0],
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            dim=dim,
            nb_batch_per_step=nb_batch_per_step,
        )
        self.add_module("Concat", blocks.Concat(), in_branch=[0, 1], out_branch=["input_concat"])
        self.add_module(
            "UNetBlock_0",
            UNet.UNetBlock(
                channels,
                nb_conv_per_stage,
                block_config,
                downsample_mode=blocks.DownsampleMode[downsample_mode],
                upsample_mode=blocks.UpsampleMode[upsample_mode],
                attention=attention,
                block=blocks.ConvBlock,
                nb_class=0,
                dim=dim,
            ),
            in_branch=["input_concat"],
            out_branch=["unet"],
        )

        if rigid:
            self.add_module(
                "Flow",
                Rigid(channels[1], dim),
                in_branch=["unet"],
                out_branch=["pos_flow"],
            )
        else:
            self.add_module(
                "Flow",
                Flow(channels[1], int_steps, int_downsize, shape, dim),
                in_branch=["unet"],
                out_branch=["pos_flow"],
            )
        self.add_module(
            "MovingImageResample",
            SpatialTransformer(shape, rigid=rigid),
            in_branch=[1, "pos_flow"],
            out_branch=["moving_image_resample"],
        )


class Flow(network.ModuleArgsDict):

    def __init__(
        self,
        in_channels: int,
        int_steps: int,
        int_downsize: int,
        shape: list[int],
        dim: int,
    ) -> None:
        super().__init__()
        self.add_module(
            "Head",
            blocks.get_torch_module("Conv", dim)(
                in_channels=in_channels,
                out_channels=dim,
                kernel_size=3,
                stride=1,
                padding=1,
            ),
        )
        self["Head"].weight = Parameter(torch.distributions.Normal(0, 1e-5).sample(self["Head"].weight.shape))
        self["Head"].bias = Parameter(torch.zeros(self["Head"].bias.shape))

        if int_steps > 0 and int_downsize > 1:
            self.add_module("DownSample", ResizeTransform(int_downsize))

        if int_steps > 0:
            self.add_module(
                "Integrate_pos_flow",
                VecInt([int(dim / int_downsize) for dim in shape], int_steps),
            )

        if int_steps > 0 and int_downsize > 1:
            self.add_module("Upsample_pos_flow", ResizeTransform(1 / int_downsize))


class Rigid(network.ModuleArgsDict):

    def __init__(self, in_channels: int, dim: int) -> None:
        super().__init__()
        self.add_module("ToFeatures", torch.nn.Flatten(1))
        self.add_module("Head", torch.nn.Linear(in_channels * 512 * 512, 2))

    def init(self, init_type: str, init_gain: float):
        self["Head"].weight.data.fill_(0)
        self["Head"].bias.data.copy_(torch.tensor([0, 0], dtype=torch.float))


class MaskFlow(torch.nn.Module):

    def __init__(self):
        super().__init__()

    def forward(self, mask: torch.Tensor, *flows: torch.Tensor):
        result = torch.zeros_like(flows[0])
        for i, flow in enumerate(flows):
            result = result + torch.where(mask == i + 1, flow, torch.tensor(0))
        return result


class SpatialTransformer(torch.nn.Module):

    def __init__(self, size: list[int], rigid: bool = False):
        super().__init__()
        self.rigid = rigid
        if not rigid:
            vectors = [torch.arange(0, s) for s in size]
            grids = torch.meshgrid(vectors, indexing="ij")
            grid = torch.stack(grids)
            grid = torch.unsqueeze(grid, 0)
            grid = grid.type(torch.float)
            self.register_buffer("grid", grid)

    def forward(self, src: torch.Tensor, flow: torch.Tensor):
        if self.rigid:
            new_locs = torch.zeros((flow.shape[0], 2, 3)).to(flow.device)
            new_locs[:, 0, 0] = 1
            new_locs[:, 1, 1] = 1
            new_locs[:, 0, 2] = flow[:, 0]
            new_locs[:, 1, 2] = flow[:, 1]
            return F.grid_sample(
                src,
                F.affine_grid(new_locs, src.size()),
                align_corners=True,
                mode="bilinear",
            )
        else:
            new_locs = self.grid + flow
            shape = flow.shape[2:]
            for i in range(len(shape)):
                new_locs[:, i, ...] = 2 * (new_locs[:, i, ...] / (shape[i] - 1) - 0.5)
            new_locs = new_locs.permute(0, 2, 3, 1)
            return F.grid_sample(src, new_locs[..., [1, 0]], align_corners=True, mode="bilinear")


class VecInt(torch.nn.Module):

    def __init__(self, inshape: list[int], nsteps: int):
        super().__init__()
        if nsteps < 0:
            raise NameError(f"nsteps should be >= 0, found: {nsteps}")
        self.nsteps = nsteps
        self.scale = 1.0 / (2**self.nsteps)
        self.transformer = SpatialTransformer(inshape)

    def forward(self, vec: torch.Tensor):
        vec = vec * self.scale
        for _ in range(self.nsteps):
            vec = vec + self.transformer(vec, vec)
        return vec


class ResizeTransform(torch.nn.Module):

    def __init__(self, size: float):
        super().__init__()
        self.factor = 1.0 / size

    def forward(self, x: torch.Tensor):
        if self.factor < 1:
            x = F.interpolate(
                x,
                align_corners=True,
                scale_factor=self.factor,
                mode="bilinear",
                recompute_scale_factor=True,
            )
            x = self.factor * x
        elif self.factor > 1:
            x = self.factor * x
            x = F.interpolate(
                x,
                align_corners=True,
                scale_factor=self.factor,
                mode="bilinear",
                recompute_scale_factor=True,
            )
        return x



import torch

from konfai.data.patching import ModelPatch
from konfai.network import blocks, network


class UNetHead(network.ModuleArgsDict):

    def __init__(self, in_channels: int, nb_class: int, dim: int, level: int) -> None:
        super().__init__()
        self.add_module(
            "Conv",
            blocks.get_torch_module("Conv", dim)(
                in_channels=in_channels,
                out_channels=nb_class,
                kernel_size=1,
                stride=1,
                padding=0,
            ),
        )
        self.add_module("Softmax", torch.nn.Softmax(dim=1))
        self.add_module("Argmax", blocks.ArgMax(dim=1))


class UNetBlock(network.ModuleArgsDict):

    def __init__(
        self,
        channels: list[int],
        nb_conv_per_stage: int,
        block_config: blocks.BlockConfig,
        downsample_mode: blocks.DownsampleMode,
        upsample_mode: blocks.UpsampleMode,
        attention: bool,
        block: type,
        nb_class: int,
        dim: int,
        i: int = 0,
    ) -> None:
        super().__init__()
        block_config_stride = block_config
        if i > 0:
            if downsample_mode != blocks.DownsampleMode.CONV_STRIDE:
                self.add_module(
                    downsample_mode.name,
                    blocks.downsample(
                        in_channels=channels[0],
                        out_channels=channels[1],
                        downsample_mode=downsample_mode,
                        dim=dim,
                    ),
                )
            else:
                block_config_stride = blocks.BlockConfig(
                    block_config.kernel_size,
                    2,
                    block_config.padding,
                    block_config.bias,
                    block_config.activation,
                    block_config.norm_mode,
                )
        self.add_module(
            "DownConvBlock",
            block(
                in_channels=channels[0],
                out_channels=channels[1],
                block_configs=[block_config_stride] + [block_config] * (nb_conv_per_stage - 1),
                dim=dim,
            ),
        )
        if len(channels) > 2:
            self.add_module(
                f"UNetBlock_{i + 1}",
                UNetBlock(
                    channels[1:],
                    nb_conv_per_stage,
                    block_config,
                    downsample_mode,
                    upsample_mode,
                    attention,
                    block,
                    nb_class,
                    dim,
                    i + 1,
                ),
            )
            self.add_module(
                "UpConvBlock",
                block(
                    in_channels=(
                        (channels[1] + channels[2])
                        if upsample_mode != blocks.UpsampleMode.CONV_TRANSPOSE
                        else channels[1] * 2
                    ),
                    out_channels=channels[1],
                    block_configs=[block_config] * nb_conv_per_stage,
                    dim=dim,
                ),
            )
            if nb_class > 0:
                self.add_module("Head", UNetHead(channels[1], nb_class, dim, i), out_branch=[-1])
        if i > 0:
            if attention:
                self.add_module(
                    "Attention",
                    blocks.Attention(f_g=channels[1], f_l=channels[0], f_int=channels[0], dim=dim),
                    in_branch=[1, 0],
                    out_branch=[1],
                )
            self.add_module(
                upsample_mode.name,
                blocks.upsample(
                    in_channels=channels[1],
                    out_channels=channels[0],
                    upsample_mode=upsample_mode,
                    dim=dim,
                    kernel_size=2,
                    stride=2,
                ),
            )
            self.add_module("SkipConnection", blocks.Concat(), in_branch=[0, 1])


class UNet(network.Network):

    def __init__(
        self,
        optimizer: network.OptimizerLoader = network.OptimizerLoader(),
        schedulers: dict[str, network.LRSchedulersLoader] = {
            "default:ReduceLROnPlateau": network.LRSchedulersLoader(0)
        },
        outputs_criterions: dict[str, network.TargetCriterionsLoader] = {"default": network.TargetCriterionsLoader()},
        patch: ModelPatch | None = None,
        dim: int = 3,
        channels: list[int] = [1, 64, 128, 256, 512, 1024],
        nb_class: int = 2,
        block_config: blocks.BlockConfig = blocks.BlockConfig(),
        nb_conv_per_stage: int = 2,
        downsample_mode: str = "MAXPOOL",
        upsample_mode: str = "CONV_TRANSPOSE",
        attention: bool = False,
        block_type: str = "Conv",
    ) -> None:
        super().__init__(
            in_channels=channels[0],
            optimizer=optimizer,
            schedulers=schedulers,
            outputs_criterions=outputs_criterions,
            patch=patch,
            dim=dim,
        )
        self.add_module(
            "UNetBlock_0",
            UNetBlock(
                channels,
                nb_conv_per_stage,
                block_config,
                downsample_mode=blocks.DownsampleMode[downsample_mode],
                upsample_mode=blocks.UpsampleMode[upsample_mode],
                attention=attention,
                block=blocks.ConvBlock if block_type == "Conv" else blocks.ResBlock,
                nb_class=nb_class,
                dim=dim,
            ),
        )



import numpy as np
import scipy
import SimpleITK as sitk  # noqa: N813
import torch
import torch.nn.functional as F  # noqa: N812

from konfai.utils.utils import _resample


def _open_transform(
    transform_files: dict[str | sitk.Transform, bool], image: sitk.Image = None
) -> list[sitk.Transform]:
    transforms: list[sitk.Transform] = []

    for transform_file, invert in transform_files.items():
        if isinstance(transform_file, str):
            transform = sitk.ReadTransform(transform_file + ".itk.txt")
        else:
            transform = transform_file
        if transform.GetName() == "TranslationTransform":
            transform = sitk.TranslationTransform(transform)
            if invert:
                transform = sitk.TranslationTransform(transform.GetInverse())
        elif transform.GetName() == "Euler3DTransform":
            transform = sitk.Euler3DTransform(transform)
            if invert:
                transform = sitk.Euler3DTransform(transform.GetInverse())
        elif transform.GetName() == "VersorRigid3DTransform":
            transform = sitk.VersorRigid3DTransform(transform)
            if invert:
                transform = sitk.VersorRigid3DTransform(transform.GetInverse())
        elif transform.GetName() == "AffineTransform":
            transform = sitk.AffineTransform(transform)
            if invert:
                transform = sitk.AffineTransform(transform.GetInverse())
        elif transform.GetName() == "DisplacementFieldTransform":
            if invert:
                transform_to_displacement_field_filter = sitk.TransformToDisplacementFieldFilter()
                transform_to_displacement_field_filter.SetReferenceImage(image)
                displacement_field = transform_to_displacement_field_filter.Execute(transform)
                iterative_inverse_displacement_field_image_filter = sitk.IterativeInverseDisplacementFieldImageFilter()
                iterative_inverse_displacement_field_image_filter.SetNumberOfIterations(20)
                inverse_displacement_field = iterative_inverse_displacement_field_image_filter.Execute(
                    displacement_field
                )
                transform = sitk.DisplacementFieldTransform(inverse_displacement_field)
            transforms.append(transform)
        else:
            transform = sitk.BSplineTransform(transform)
            if invert:
                transform_to_displacement_field_filter = sitk.TransformToDisplacementFieldFilter()
                transform_to_displacement_field_filter.SetReferenceImage(image)
                displacement_field = transform_to_displacement_field_filter.Execute(transform)
                iterative_inverse_displacement_field_image_filter = sitk.IterativeInverseDisplacementFieldImageFilter()
                iterative_inverse_displacement_field_image_filter.SetNumberOfIterations(20)
                inverse_displacement_field = iterative_inverse_displacement_field_image_filter.Execute(
                    displacement_field
                )
                transform = sitk.DisplacementFieldTransform(inverse_displacement_field)
        transforms.append(transform)
    if len(transforms) == 0:
        transforms.append(sitk.Euler3DTransform())
    return transforms


def _open_rigid_transform(transform_files: dict[str | sitk.Transform, bool]) -> tuple[np.ndarray, np.ndarray]:
    transforms = _open_transform(transform_files)
    matrix_result = np.identity(3)
    translation_result = np.array([0, 0, 0])

    for transform in transforms:
        if hasattr(transform, "GetMatrix"):
            matrix = np.linalg.inv(np.array(transform.GetMatrix(), dtype=np.double).reshape((3, 3)))
            translation = -np.asarray(transform.GetTranslation(), dtype=np.double)
            center = np.asarray(transform.GetCenter(), dtype=np.double)
        else:
            matrix = np.eye(len(transform.GetOffset()))
            translation = -np.asarray(transform.GetOffset(), dtype=np.double)
            center = np.asarray([0] * len(transform.GetOffset()), dtype=np.double)

        translation_center = np.linalg.inv(matrix).dot(matrix.dot(translation - center) + center)
        translation_result = np.linalg.inv(matrix_result).dot(translation_center) + translation_result
        matrix_result = matrix.dot(matrix_result)
    return np.linalg.inv(matrix_result), -translation_result


def compose_transform(
    transform_files: dict[str | sitk.Transform, bool], image: sitk.Image = None
) -> sitk.CompositeTransform:
    transforms = _open_transform(transform_files, image)
    result = sitk.CompositeTransform(transforms)
    return result


def flatten_transform(transform_files: dict[str | sitk.Transform, bool]) -> sitk.AffineTransform:
    [matrix, translation] = _open_rigid_transform(transform_files)
    transform = sitk.AffineTransform(3)
    transform.SetMatrix(matrix.flatten())
    transform.SetTranslation(translation)
    return transform


def apply_to_image_rigid_transform(image: sitk.Image, transform_files: dict[str | sitk.Transform, bool]) -> sitk.Image:
    [matrix, translation] = _open_rigid_transform(transform_files)
    matrix = np.linalg.inv(matrix)
    translation = -translation
    data = sitk.GetArrayFromImage(image)
    result = sitk.GetImageFromArray(data)
    result.SetDirection(matrix.dot(np.array(image.GetDirection()).reshape((3, 3))).flatten())
    result.SetOrigin(matrix.dot(np.array(image.GetOrigin()) + translation))
    result.SetSpacing(image.GetSpacing())
    return result


def apply_to_data_transform(data: np.ndarray, transform_files: dict[str | sitk.Transform, bool]) -> sitk.Image:
    transforms = compose_transform(transform_files)
    result = np.copy(data)
    # _LPS = lambda matrix: np.array([-matrix[0], -matrix[1], matrix[2]], dtype=np.double)
    for i in range(data.shape[0]):
        result[i, :] = transforms.TransformPoint(np.asarray(data[i, :], dtype=np.double))
    return result


def resample_itk(
    image_reference: sitk.Image,
    image: sitk.Image,
    transform_files: dict[str | sitk.Transform, bool],
    mask=False,
    default_pixel_value: float | None = None,
    torch_resample: bool = False,
) -> sitk.Image:
    if torch_resample:
        input_tensor = torch.tensor(sitk.GetArrayFromImage(image)).unsqueeze(0)
        vectors = [torch.arange(0, s) for s in input_tensor.shape[1:]]
        grids = torch.meshgrid(vectors, indexing="ij")
        grid = torch.stack(grids)
        grid = torch.unsqueeze(grid, 0)
        transform_to_displacement_field_filter = sitk.TransformToDisplacementFieldFilter()
        transform_to_displacement_field_filter.SetReferenceImage(image)
        transform_to_displacement_field_filter.SetNumberOfThreads(16)
        new_locs = grid + torch.tensor(
            sitk.GetArrayFromImage(
                transform_to_displacement_field_filter.Execute(compose_transform(transform_files, image))
            )
        ).unsqueeze(0).permute(0, 4, 1, 2, 3)
        shape = new_locs.shape[2:]
        for i in range(len(shape)):
            new_locs[:, i, ...] = 2 * (new_locs[:, i, ...] / (shape[i] - 1) - 0.5)
        new_locs = new_locs.permute(0, 2, 3, 4, 1)
        new_locs = new_locs[..., [2, 1, 0]]
        result_data = F.grid_sample(
            input_tensor.unsqueeze(0).float(),
            new_locs.float(),
            align_corners=True,
            padding_mode="border",
            mode="nearest" if input_tensor.dtype == torch.uint8 else "bilinear",
        ).squeeze(0)
        result_data = result_data.type(torch.uint8) if input_tensor.dtype == torch.uint8 else result_data
        result = sitk.GetImageFromArray(result_data.squeeze(0).numpy())
        result.CopyInformation(image_reference)
        return result
    else:
        return sitk.Resample(
            image,
            image_reference,
            compose_transform(transform_files, image),
            sitk.sitkNearestNeighbor if mask else sitk.sitkBSpline,
            (
                default_pixel_value
                if default_pixel_value is not None
                else (0 if mask else int(np.min(sitk.GetArrayFromImage(image))))
            ),
        )


def parametermap_to_transform(
    path_src: str,
) -> sitk.Transform | list[sitk.Transform]:
    transform = sitk.ReadParameterFile(path_src)

    def array_format(x):
        return [float(i) for i in x]

    dimension = int(transform["FixedImageDimension"][0])

    if transform["Transform"][0] == "EulerTransform":
        if dimension == 2:
            result = sitk.Euler2DTransform()
        else:
            result = sitk.Euler3DTransform()
        parameters = array_format(transform["TransformParameters"])
        fixed_parameters = array_format(transform["CenterOfRotationPoint"]) + [0]
    elif transform["Transform"][0] == "TranslationTransform":
        result = sitk.TranslationTransform(dimension)
        parameters = array_format(transform["TransformParameters"])
        fixed_parameters = []
    elif transform["Transform"][0] == "AffineTransform":
        result = sitk.AffineTransform(dimension)
        parameters = array_format(transform["TransformParameters"])
        fixed_parameters = array_format(transform["CenterOfRotationPoint"]) + [0]
    elif transform["Transform"][0] == "BSplineStackTransform":
        parameters = array_format(transform["TransformParameters"])
        grid_size = array_format(transform["GridSize"])
        grid_origin = array_format(transform["GridOrigin"])
        grid_spacing = array_format(transform["GridSpacing"])
        grid_direction = (
            np.asarray(array_format(transform["GridDirection"])).reshape((dimension, dimension)).T.flatten()
        )
        fixed_parameters = np.concatenate([grid_size, grid_origin, grid_spacing, grid_direction])

        nb = int(array_format(transform["Size"])[-1])
        sub = int(np.prod(grid_size)) * dimension
        results = []
        for i in range(nb):
            result = sitk.BSplineTransform(dimension)
            sub_parameters = np.asarray(parameters[i * sub : (i + 1) * sub])
            result.SetFixedParameters(fixed_parameters)
            result.SetParameters(sub_parameters)
            results.append(result)
        return results
    elif transform["Transform"][0] == "AffineLogStackTransform":
        parameters = array_format(transform["TransformParameters"])
        fixed_parameters = array_format(transform["CenterOfRotationPoint"]) + [0]

        nb = int(transform["NumberOfSubTransforms"][0])
        sub = dimension * 4
        results = []
        for i in range(nb):
            result = sitk.AffineTransform(dimension)
            sub_parameters = np.asarray(parameters[i * sub : (i + 1) * sub])

            result.SetFixedParameters(fixed_parameters)
            result.SetParameters(
                np.concatenate(
                    [
                        scipy.linalg.expm(
                            sub_parameters[: dimension * dimension].reshape((dimension, dimension))
                        ).flatten(),
                        sub_parameters[-dimension:],
                    ]
                )
            )
            results.append(result)
        return results
    elif transform["Transform"][0] == "BSplineTransform":
        result = sitk.BSplineTransform(dimension)

        parameters = array_format(transform["TransformParameters"])
        grid_size = array_format(transform["GridSize"])
        grid_origin = array_format(transform["GridOrigin"])
        grid_spacing = array_format(transform["GridSpacing"])
        grid_direction = np.array(array_format(transform["GridDirection"])).reshape((dimension, dimension)).T.flatten()
        fixed_parameters = np.concatenate([grid_size, grid_origin, grid_spacing, grid_direction])
    else:
        raise NameError(f"Transform {transform['Transform'][0]} doesn't exist")
    result.SetFixedParameters(fixed_parameters)
    result.SetParameters(parameters)
    return result


def resample_isotropic(image: sitk.Image, spacing: list[float] | None = None) -> sitk.Image:
    spacing = spacing or [1.0, 1.0, 1.0]
    resize_factor = [y / x for x, y in zip(spacing, image.GetSpacing())]
    result = sitk.GetImageFromArray(
        _resample(
            torch.tensor(sitk.GetArrayFromImage(image)).unsqueeze(0),
            [int(size * factor) for size, factor in zip(image.GetSize(), resize_factor)],
        )
        .squeeze(0)
        .numpy()
    )
    result.SetDirection(image.GetDirection())
    result.SetOrigin(image.GetOrigin())
    result.SetSpacing(spacing)
    return result


def resample_resize(image: sitk.Image, size: list[int] | None = None):
    size = size or [100, 512, 512]
    result = sitk.GetImageFromArray(
        _resample(torch.tensor(sitk.GetArrayFromImage(image)).unsqueeze(0), size).squeeze(0).numpy()
    )
    result.SetDirection(image.GetDirection())
    result.SetOrigin(image.GetOrigin())
    result.SetSpacing([x / y * z for x, y, z in zip(image.GetSize(), size, image.GetSpacing())])
    return result


def box_with_mask(mask: sitk.Image, label: list[int], dilatations: list[int]) -> np.ndarray:

    dilatations = [int(np.ceil(d / s)) for d, s in zip(dilatations, reversed(mask.GetSpacing()))]

    data = sitk.GetArrayFromImage(mask)
    border = np.where(np.isin(sitk.GetArrayFromImage(mask), label))
    box = []
    for w, dilatation, s in zip(border, dilatations, data.shape):
        box.append([max(np.min(w) - dilatation, 0), min(np.max(w) + dilatation, s)])
    box = np.asarray(box)
    return box


def crop_with_mask(image: sitk.Image, box: np.ndarray) -> sitk.Image:
    data = sitk.GetArrayFromImage(image)

    for i, w in enumerate(box):
        data = np.delete(data, slice(w[1], data.shape[i]), i)
        data = np.delete(data, slice(0, w[0]), i)

    origin = np.asarray(image.GetOrigin())
    matrix = np.asarray(image.GetDirection()).reshape((len(origin), len(origin)))
    origin = origin.dot(matrix)
    for i, w in enumerate(box):
        origin[-i - 1] += w[0] * np.asarray(image.GetSpacing())[-i - 1]
    origin = origin.dot(np.linalg.inv(matrix))

    result = sitk.GetImageFromArray(data)
    result.SetOrigin(origin)
    result.SetSpacing(image.GetSpacing())
    result.SetDirection(image.GetDirection())
    return result


def format_mask_label(mask: sitk.Image, labels: list[tuple[int, int]]) -> sitk.Image:
    data = sitk.GetArrayFromImage(mask)
    result_data = np.zeros_like(data, np.uint8)

    for label_old, label_new in labels:
        result_data[np.where(data == label_old)] = label_new

    result = sitk.GetImageFromArray(result_data)
    result.CopyInformation(mask)
    return result


def get_flat_label(mask: sitk.Image, labels: None | list[int] = None) -> sitk.Image:
    data = sitk.GetArrayFromImage(mask)
    result_data = np.zeros_like(data, np.uint8)
    if labels is not None:
        for label in labels:
            result_data[np.where(data == label)] = 1
    else:
        result_data[np.where(data > 0)] = 1
    result = sitk.GetImageFromArray(result_data)
    result.CopyInformation(mask)
    return result


def clip_and_cast(image: sitk.Image, min_value: float, max_value: float, dtype: np.dtype) -> sitk.Image:
    data = sitk.GetArrayFromImage(image)
    data[np.where(data > max_value)] = max_value
    data[np.where(data < min_value)] = min_value
    result = sitk.GetImageFromArray(data.astype(dtype))
    result.CopyInformation(image)
    return result



import os
import numpy as np
import pytest
import shutil
import tempfile
import SimpleITK as sitk

from konfai.utils.dataset import Dataset, Attribute, data_to_image, image_to_data


@pytest.fixture
def temp_dataset_dir():
    temp_dir = tempfile.mkdtemp()
    yield temp_dir
    shutil.rmtree(temp_dir)

def create_dummy_image():
    img = sitk.GaussianSource(size=[32, 32, 32], sigma=[1.0]*3)
    return img


def test_write_and_read_image(temp_dataset_dir):
    dataset = Dataset(temp_dataset_dir, format="npy")
    image = create_dummy_image()
    attr = Attribute()
    data, attrs = image_to_data(image)
    attr.update(attrs)

    dataset.write("GroupA", "TestImage", image, attr)
    image_out = dataset.readImage("GroupA", "TestImage")

    assert isinstance(image_out, sitk.Image)
    assert image_out.GetSize() == image.GetSize()


def test_get_names_and_groups(temp_dataset_dir):
    dataset = Dataset(temp_dataset_dir, format="npy")
    image = create_dummy_image()
    attr = Attribute()
    data, attrs = image_to_data(image)
    attr.update(attrs)

    dataset.write("GroupA", "Test1", image, attr)
    dataset.write("GroupA", "Test2", image, attr)

    names = dataset.getNames("GroupA")
    groups = dataset.getGroup()

    assert "Test1" in names
    assert "Test2" in names
    assert any("GroupA" in g for g in groups)


def test_dataset_exists(temp_dataset_dir):
    dataset = Dataset(temp_dataset_dir, format="npy")
    image = create_dummy_image()
    attr = Attribute()
    data, attrs = image_to_data(image)
    attr.update(attrs)

    dataset.write("GroupA", "ExistsImage", image, attr)

    assert dataset.isDatasetExist("GroupA", "ExistsImage")
    assert not dataset.isDatasetExist("GroupA", "MissingImage")


def test_read_data_and_transform(temp_dataset_dir):
    dataset = Dataset(temp_dataset_dir, format="npy")
    image = create_dummy_image()
    attr = Attribute()
    data, attrs = image_to_data(image)
    attr.update(attrs)

    dataset.write("GroupA", "ImageData", data, attr)

    read_data, read_attr = dataset.readData("GroupA", "ImageData")

    assert isinstance(read_data, np.ndarray)
    assert "Origin" in read_attr


import os
import shutil

import torch
import torch.distributed as dist
import tqdm
from torch.nn.parallel import DistributedDataParallel as DDP  # noqa: N817
from torch.optim.swa_utils import AveragedModel
from torch.utils.data import DataLoader
from torch.utils.tensorboard.writer import SummaryWriter

from konfai import (
    checkpoints_directory,
    config_file,
    current_date,
    konfai_state,
    path_to_models,
    setups_directory,
    statistics_directory,
)
from konfai.data.data_manager import DataTrain
from konfai.network.network import CPUModel, ModelLoader, NetState, Network
from konfai.utils.config import config
from konfai.utils.utils import DataLog, DistributedObject, State, TrainerError, description


class EarlyStoppingBase:

    def __init__(self):
        pass

    def is_stopped(self) -> bool:

        return False

    def get_score(self, values: dict[str, float]):
        return sum(list(values.values()))

    def __call__(self, current_score: float) -> bool:
        return False


class EarlyStopping(EarlyStoppingBase):
    """
    Implements early stopping logic with configurable patience and monitored metrics.

    Attributes:
        monitor (list[str]): Metrics to monitor.
        patience (int): Number of checks with no improvement before stopping.
        min_delta (float): Minimum change to qualify as improvement.
        mode (str): "min" or "max" depending on optimization direction.
    """

    @config("EarlyStopping")
    def __init__(
        self,
        monitor: list[str] | None = None,
        patience: int = 10,
        min_delta: float = 0.0,
        mode: str = "min",
    ):
        super().__init__()
        self.monitor = [] if monitor is None else monitor
        self.patience = patience
        self.min_delta = min_delta
        self.mode = mode
        self.counter = 0
        self.best_score: float | None = None
        self.early_stop = False

    def is_stopped(self) -> bool:
        return self.early_stop

    def get_score(self, values: dict[str, float]):
        if len(self.monitor) == 0:
            return super().get_score(values)
        for v in self.monitor:
            if v not in values.keys():
                raise TrainerError(
                    "Metric '{}' specified in EarlyStopping.monitor not found in logged values. ",
                    f"Available keys: {v}. Please check your configuration.",
                )
        return sum([i for v, i in values.items() if v in self.monitor])

    def __call__(self, current_score: float) -> bool:
        if self.best_score is None:
            self.best_score = current_score
            return False

        if self.mode == "min":
            improvement = self.best_score - current_score
        elif self.mode == "max":
            improvement = current_score - self.best_score
        else:
            raise TrainerError("Mode must be 'min' or 'max'.")

        if improvement > self.min_delta:
            self.best_score = current_score
            self.counter = 0
        else:
            self.counter += 1

        if self.counter >= self.patience:
            self.early_stop = True

        return self.early_stop


class _Trainer:
    """
    Internal class for managing the training loop in a distributed or standalone setting.

    Handles:
    - Epoch iteration with training and optional validation
    - Mixed precision support (autocast)
    - Exponential Moving Average (EMA) model tracking
    - Early stopping
    - Logging to TensorBoard
    - Model checkpoint saving and selection (ALL or BEST)

    This class is intended to be used via a context manager
    (`with _Trainer(...) as trainer:`)  inside the public `Trainer` class.
    """

    def __init__(
        self,
        world_size: int,
        global_rank: int,
        local_rank: int,
        size: int,
        train_name: str,
        early_stopping: EarlyStopping | None,
        data_log: list[str] | None,
        save_checkpoint_mode: str,
        epochs: int,
        epoch: int,
        autocast: bool,
        it_validation: int | None,
        it: int,
        model: DDP | CPUModel,
        model_ema: AveragedModel,
        dataloader_training: DataLoader,
        dataloader_validation: DataLoader | None = None,
    ) -> None:
        self.world_size = world_size
        self.global_rank = global_rank
        self.local_rank = local_rank
        self.size = size
        self.save_checkpoint_mode = save_checkpoint_mode
        self.train_name = train_name
        self.epochs = epochs
        self.epoch = epoch
        self.model = model
        self.dataloader_training = dataloader_training
        self.dataloader_validation = dataloader_validation
        self.autocast = autocast
        self.model_ema = model_ema
        self.early_stopping = EarlyStoppingBase() if early_stopping is None else early_stopping

        self.it_validation = len(dataloader_training) if it_validation is None else it_validation
        self.it = it
        self.tb = SummaryWriter(log_dir=statistics_directory() + self.train_name + "/")
        self.data_log: dict[str, tuple[DataLog, int]] = {}
        if data_log is not None:
            for data in data_log:
                self.data_log[data.split("/")[0].replace(":", ".")] = (
                    DataLog[data.split("/")[1]],
                    int(data.split("/")[2]),
                )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, value, traceback):
        """Closes the SummaryWriter if used."""
        if self.tb is not None:
            self.tb.close()

    def run(self) -> None:
        """
        Launches the training loop, performing one epoch at a time.
        Triggers early stopping and resets data augmentations between epochs.
        """
        self.dataloader_training.dataset.load("Train")
        if self.dataloader_validation is not None:
            self.dataloader_validation.dataset.load("Validation")
            if State[konfai_state()] != State.TRAIN:
                self._validate()

        with tqdm.tqdm(
            iterable=range(self.epoch, self.epochs),
            leave=False,
            total=self.epochs,
            initial=self.epoch,
            desc="Progress",
        ) as epoch_tqdm:
            for self.epoch in epoch_tqdm:
                self.train()
                if self.early_stopping.is_stopped():
                    break
                self.dataloader_training.dataset.reset_augmentation("Train")

    def get_input(
        self,
        data_dict: dict[str, tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, list[str], torch.Tensor]],
    ) -> dict[tuple[str, bool], torch.Tensor]:
        """
        Extracts input tensors from the data dict for model input.

        Args:
            data_dict (dict): Dictionary with data items structured as tuples.

        Returns:
            dict: Mapping from (group, bool_flag) to input tensors.
        """
        return {(k, v[5][0].item()): v[0] for k, v in data_dict.items()}

    def train(self) -> None:
        """
        Performs a full training epoch with support for:
        - mixed precision
        - DDP / CPU training
        - EMA updates
        - loss logging and checkpoint saving
        - validation at configurable iteration interval
        """
        self.model.train()
        self.model.module.set_state(NetState.TRAIN)
        if self.model_ema is not None:
            self.model_ema.eval()
            self.model_ema.module.set_state(NetState.TRAIN)

        with tqdm.tqdm(
            iterable=enumerate(self.dataloader_training),
            desc=f"Training : {description(self.model, self.model_ema)}",
            total=len(self.dataloader_training),
            leave=False,
            ncols=0,
        ) as batch_iter:
            for _, data_dict in batch_iter:
                with torch.amp.autocast("cuda", enabled=self.autocast):
                    input_data_dict = self.get_input(data_dict)
                    self.model(input_data_dict)
                    self.model.module.backward(self.model.module)
                    if self.model_ema is not None:
                        self.model_ema.update_parameters(self.model)
                        self.model_ema.module(input_data_dict)
                    self.it += 1
                    if (self.it) % self.it_validation == 0:
                        loss = self._train_log(data_dict)

                        if self.dataloader_validation is not None:
                            loss = self._validate()
                        self.model.module.update_lr()
                        score = self.early_stopping.get_score(loss)
                        self.checkpoint_save(score)
                        if self.early_stopping(score):
                            break

                batch_iter.set_description(f"Training : {description(self.model, self.model_ema)}")

    @torch.no_grad()
    def _validate(self) -> float:
        """
        Executes the validation phase, evaluates loss and metrics.
        Updates model states and resets augmentation for validation set.

        Returns:
            float: Validation loss.
        """
        if self.dataloader_validation is None:
            return 0
        self.model.eval()
        self.model.module.set_state(NetState.PREDICTION)
        if self.model_ema is not None:
            self.model_ema.module.set_state(NetState.PREDICTION)

        data_dict = None
        with tqdm.tqdm(
            iterable=enumerate(self.dataloader_validation),
            desc=f"Validation : {description(self.model, self.model_ema)}",
            total=len(self.dataloader_validation),
            leave=False,
            ncols=0,
        ) as batch_iter:
            for _, data_dict in batch_iter:
                input_data_dict = self.get_input(data_dict)
                self.model(input_data_dict)
                if self.model_ema is not None:
                    self.model_ema.module(input_data_dict)

                batch_iter.set_description(f"Validation : {description(self.model, self.model_ema)}")
        self.dataloader_validation.dataset.reset_augmentation("Validation")
        dist.barrier()
        self.model.train()
        self.model.module.set_state(NetState.TRAIN)
        if self.model_ema is not None:
            self.model_ema.module.set_state(NetState.TRAIN)
        return self._validation_log(data_dict)

    def checkpoint_save(self, loss: float) -> None:
        """
        Saves model and optimizer states. Keeps either all checkpoints or only the best one.

        Args:
            loss (float): Current loss used for best checkpoint selection.
        """
        if self.global_rank != 0:
            return

        path = checkpoints_directory() + self.train_name + "/"
        os.makedirs(path, exist_ok=True)

        name = current_date() + ".pt"
        save_path = os.path.join(path, name)

        save_dict = {
            "epoch": self.epoch,
            "it": self.it,
            "loss": loss,
            "Model": self.model.module.state_dict(),
        }

        if self.model_ema is not None:
            save_dict["Model_EMA"] = self.model_ema.module.state_dict()

        save_dict.update(
            {
                f"{name}_optimizer_state_dict": network.optimizer.state_dict()
                for name, network in self.model.module.get_networks().items()
                if network.optimizer is not None
            }
        )
        save_dict.update(
            {
                f"{name}_it": network._it
                for name, network in self.model.module.get_networks().items()
                if network.optimizer is not None
            }
        )
        save_dict.update(
            {
                f"{name}_nb_lr_update": network._nb_lr_update
                for name, network in self.model.module.get_networks().items()
                if network.optimizer is not None
            }
        )

        torch.save(save_dict, save_path)

        if self.save_checkpoint_mode == "BEST":
            all_checkpoints = sorted([os.path.join(path, f) for f in os.listdir(path) if f.endswith(".pt")])
            best_ckpt = None
            best_loss = float("inf")

            for f in all_checkpoints:
                d = torch.load(f, weights_only=True)
                if d.get("loss", float("inf")) < best_loss:
                    best_loss = d["loss"]
                    best_ckpt = f

            for f in all_checkpoints:
                if f != best_ckpt and f != save_path:
                    os.remove(f)

    @torch.no_grad()
    def _log(
        self,
        type_log: str,
        data_dict: dict[str, tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, list[str], torch.Tensor]],
    ) -> dict[str, float] | None:
        """
        Logs losses, metrics and optionally images to TensorBoard.

        Args:
            type_log (str): "Training" or "Validation".
            data_dict (dict): Dictionary of data from current batch.

        Returns:
            dict[str, float] | None: Dictionary of aggregated losses and metrics if rank == 0.
        """
        models: dict[str, Network] = {"": self.model.module}
        if self.model_ema is not None:
            models["_EMA"] = self.model_ema.module

        measures = DistributedObject.get_measure(
            self.world_size,
            self.global_rank,
            self.local_rank * self.size + self.size - 1,
            models,
            (
                self.it_validation
                if type_log == "Training" or self.dataloader_validation is None
                else len(self.dataloader_validation)
            ),
        )

        if self.global_rank == 0:
            images_log = []
            if len(self.data_log):
                for name, data_type in self.data_log.items():
                    if name in data_dict:
                        data_type[0](
                            self.tb,
                            f"{type_log}/{name}",
                            data_dict[name][0][: self.data_log[name][1]].detach().cpu().numpy(),
                            self.it,
                        )
                    else:
                        images_log.append(name.replace(":", "."))

            for label, model in models.items():
                for name, network in model.get_networks().items():
                    if network.measure is not None:
                        self.tb.add_scalars(
                            f"{type_log}/{name}/Loss/{label}",
                            {k: v[1] for k, v in measures[f"{name}{label}"][0].items()},
                            self.it,
                        )
                        self.tb.add_scalars(
                            f"{type_log}/{name}/Loss_weight/{label}",
                            {k: v[0] for k, v in measures[f"{name}{label}"][0].items()},
                            self.it,
                        )

                        self.tb.add_scalars(
                            f"{type_log}/{name}/Metric/{label}",
                            {k: v[1] for k, v in measures[f"{name}{label}"][1].items()},
                            self.it,
                        )
                        self.tb.add_scalars(
                            f"{type_log}/{name}/Metric_weight/{label}",
                            {k: v[0] for k, v in measures[f"{name}{label}"][1].items()},
                            self.it,
                        )

                    if len(images_log):
                        for name, layer, _ in model.get_layers(
                            [v.to(0) for k, v in self.get_input(data_dict).items() if k[1]],
                            images_log,
                        ):
                            self.data_log[name][0](
                                self.tb,
                                f"{type_log}/{name}{label}",
                                layer[: self.data_log[name][1]].detach().cpu().numpy(),
                                self.it,
                            )

            if type_log == "Training":
                for name, network in self.model.module.get_networks().items():
                    if network.optimizer is not None:
                        self.tb.add_scalar(
                            f"{type_log}/{name}/Learning Rate",
                            network.optimizer.param_groups[0]["lr"],
                            self.it,
                        )

        if self.global_rank == 0:
            loss = {}
            for name, network in self.model.module.get_networks().items():
                if network.measure is not None:
                    loss.update({k: v[1] for k, v in measures[f"{name}{label}"][0].items()})
                    loss.update({k: v[1] for k, v in measures[f"{name}{label}"][1].items()})
            return loss
        return None

    @torch.no_grad()
    def _train_log(self, data_dict: dict[str, tuple[torch.Tensor, int, int, int]]) -> dict[str, float]:
        """Wrapper for _log during training."""
        return self._log("Training", data_dict)

    @torch.no_grad()
    def _validation_log(self, data_dict: dict[str, tuple[torch.Tensor, int, int, int]]) -> dict[str, float]:
        """Wrapper for _log during validation."""
        return self._log("Validation", data_dict)


class Trainer(DistributedObject):
    """
    Public API for training a model using the KonfAI framework.
    Wraps setup, checkpointing, resuming, logging, and launching distributed _Trainer.

    Main responsibilities:
    - Initialization from config (via @config)
    - Model and EMA setup
    - Checkpoint loading and saving
    - Distributed setup and launch

    Args:
        model (ModelLoader): Loader for model architecture.
        dataset (DataTrain): Training/validation dataset.
        train_name (str): Training session name.
        manual_seed (int | None): Random seed.
        epochs (int): Number of epochs to run.
        it_validation (int | None): Validation interval.
        autocast (bool): Enable AMP training.
        gradient_checkpoints (list[str] | None): Modules to use gradient checkpointing on.
        gpu_checkpoints (list[str] | None): Modules to pin on specific GPUs.
        ema_decay (float): EMA decay factor.
        data_log (list[str] | None): Logging instructions.
        early_stopping (EarlyStopping | None): Optional early stopping config.
        save_checkpoint_mode (str): Either "BEST" or "ALL".
    """

    @config("Trainer")
    def __init__(
        self,
        model: ModelLoader = ModelLoader(),
        dataset: DataTrain = DataTrain(),
        train_name: str = "default:TRAIN_01",
        manual_seed: int | None = None,
        epochs: int = 100,
        it_validation: int | None = None,
        autocast: bool = False,
        gradient_checkpoints: list[str] | None = None,
        gpu_checkpoints: list[str] | None = None,
        ema_decay: float = 0,
        data_log: list[str] | None = None,
        early_stopping: EarlyStopping | None = None,
        save_checkpoint_mode: str = "BEST",
    ) -> None:
        if os.environ["KONFAI_CONFIG_MODE"] != "Done":
            exit(0)
        super().__init__(train_name)
        self.manual_seed = manual_seed
        self.dataset = dataset
        self.autocast = autocast
        self.epochs = epochs
        self.epoch = 0
        self.early_stopping = early_stopping
        self.it = 0
        self.it_validation = it_validation
        self.model = model.get_model(train=True)
        self.ema_decay = ema_decay
        self.model_ema: torch.optim.swa_utils.AveragedModel | None = None
        self.data_log = data_log

        modules = []
        for i, _ in self.model.named_modules():
            modules.append(i)
        if self.data_log is not None:
            for k in self.data_log:
                tmp = k.split("/")[0].replace(":", ".")
                if tmp not in self.dataset.get_groups_dest() and tmp not in modules:
                    raise TrainerError(
                        f"Invalid key '{tmp}' in `data_log`.",
                        f"This key is neither a destination group from the dataset ({self.dataset.get_groups_dest()})",
                        f"nor a valid module name in the model ({modules}).",
                        "Please check your `data_log` configuration,"
                        " it should reference either a model output or a dataset group.",
                    )

        self.gradient_checkpoints = gradient_checkpoints
        self.gpu_checkpoints = gpu_checkpoints
        self.save_checkpoint_mode = save_checkpoint_mode
        self.config_namefile_src = config_file().replace(".yml", "")
        self.config_namefile = (
            setups_directory() + self.name + "/" + self.config_namefile_src.split("/")[-1] + "_" + str(self.it) + ".yml"
        )
        self.size = len(self.gpu_checkpoints) + 1 if self.gpu_checkpoints else 1

    def __exit__(self, exc_type, value, traceback):
        """Exit training context and trigger save of model/checkpoints."""
        super().__exit__(exc_type, value, traceback)
        self._save()

    def _load(self) -> dict[str, dict[str, torch.Tensor]]:
        """
        Loads a previously saved checkpoint from local disk or URL.

        Returns:
            dict: State dictionary loaded from checkpoint.
        """
        if path_to_models().startswith("https://"):
            try:
                state_dict = {
                    path_to_models().split(":")[1]: torch.hub.load_state_dict_from_url(
                        url=path_to_models().split(":")[0], map_location="cpu", check_hash=True
                    )
                }
            except Exception:
                raise Exception(f"Model : {path_to_models()} does not exist !")
        else:
            if path_to_models() != "":
                path = ""
                name = path_to_models()
            else:
                path = checkpoints_directory() + self.name + "/"
                if os.listdir(path):
                    name = sorted(os.listdir(path))[-1]

            if os.path.exists(path + name):
                state_dict = torch.load(path + name, weights_only=True, map_location="cpu")
            else:
                raise Exception(f"Model : {self.name} does not exist !")

        if "epoch" in state_dict:
            self.epoch = state_dict["epoch"]
        if "it" in state_dict:
            self.it = state_dict["it"]
        return state_dict

    def _save(self) -> None:
        """
        Serializes model and EMA model (if any) in both state_dict and full model formats.
        Also saves optimizer states and YAML config snapshot.
        """
        if os.path.exists(self.config_namefile):
            os.rename(
                self.config_namefile,
                self.config_namefile.replace(".yml", "") + "_" + str(self.it) + ".yml",
            )

    def _avg_fn(self, averaged_model_parameter: float, model_parameter, num_averaged):
        """
        EMA update rule used by AveragedModel.

        Returns:
            torch.Tensor: Blended parameter using decay factor.
        """
        return (1 - self.ema_decay) * averaged_model_parameter + self.ema_decay * model_parameter

    def setup(self, world_size: int):
        """
        Initializes the training environment:
        - Clears previous outputs (unless resuming)
        - Initializes model and EMA
        - Loads checkpoint (if resuming)
        - Prepares dataloaders

        Args:
            world_size (int): Total number of distributed processes.
        """
        state = State[konfai_state()]
        if state != State.RESUME and os.path.exists(checkpoints_directory() + self.name + "/"):
            if os.environ["KONFAI_OVERWRITE"] != "True":
                accept = input(f"The model {self.name} already exists ! Do you want to overwrite it (yes,no) : ")
                if accept != "yes":
                    exit(0)
            for directory_path in [
                statistics_directory(),
                checkpoints_directory(),
                setups_directory(),
            ]:
                if os.path.exists(directory_path + self.name + "/"):
                    shutil.rmtree(directory_path + self.name + "/")

        state_dict = {}
        if state != State.TRAIN:
            state_dict = self._load()

        self.model.init(self.autocast, state, self.dataset.get_groups_dest())
        self.model.init_outputs_group()
        self.model._compute_channels_trace(
            self.model,
            self.model.in_channels,
            self.gradient_checkpoints,
            self.gpu_checkpoints,
        )
        self.model.load(state_dict, init=True, ema=False)
        if self.ema_decay > 0:
            self.model_ema = AveragedModel(self.model, avg_fn=self._avg_fn)
            if state_dict is not None:
                self.model_ema.module.load(state_dict, init=False, ema=True)

        if not os.path.exists(setups_directory() + self.name + "/"):
            os.makedirs(setups_directory() + self.name + "/")
        shutil.copyfile(self.config_namefile_src + ".yml", self.config_namefile)

        self.dataloader, train_names, validation_names = self.dataset.get_data(world_size // self.size)
        with open(setups_directory() + self.name + "/Train_" + str(self.it) + ".txt", "w") as f:
            for name in train_names:
                f.write(name + "\n")
        with open(setups_directory() + self.name + "/Validation_" + str(self.it) + ".txt", "w") as f:
            for name in validation_names:
                f.write(name + "\n")

    def run_process(
        self,
        world_size: int,
        global_rank: int,
        local_rank: int,
        dataloaders: list[DataLoader],
    ):
        """
        Launches the actual training process via internal `_Trainer` class.
        Wraps model with DDP or CPU fallback, attaches EMA, and starts training.

        Args:
            world_size (int): Total number of distributed processes.
            global_rank (int): Global rank of the current process.
            local_rank (int): Local rank within the node.
            dataloaders (list[DataLoader]): Training and validation dataloaders.
        """
        model = Network.to(self.model, local_rank * self.size)
        model = DDP(model, static_graph=True) if torch.cuda.is_available() else CPUModel(model)
        if self.model_ema is not None:
            self.model_ema.module = Network.to(self.model_ema.module, local_rank)
        with _Trainer(
            world_size,
            global_rank,
            local_rank,
            self.size,
            self.name,
            self.early_stopping,
            self.data_log,
            self.save_checkpoint_mode,
            self.epochs,
            self.epoch,
            self.autocast,
            self.it_validation,
            self.it,
            model,
            self.model_ema,
            *dataloaders,
        ) as t:
            t.run()
